from monty.json import MSONable

from pydirac.core.periodic import Element

__all__ = ["Molecule"]


class Molecule(MSONable):
    """
    A class representing a molecule.

    Parameters
    ----------
    atoms : list, optional
        A list of atomic symbols representing the atoms in the molecule.
    coordinates : list, optional
        A list of tuples representing the cartesian coordinates of the atoms in the molecule.

    Attributes
    ----------
    atoms : list
        A list of atomic symbols representing the atoms in the molecule.
    coordinates : list
        A list of tuples representing the cartesian coordinates of the atoms in the molecule.
    is_atom : bool
        A boolean indicating whether the molecule is a single atom.
    atomic_info : str or Element
        A string representing the atomic symbol or an Element object representing the atomic properties
        of the atom in the molecule. If the molecule is not a single atom, atomic_info is set to "null".

    Methods
    -------
    as_dict() -> dict
        Returns a dictionary representation of the molecule.

    """

    def __init__(self, atoms=None, coordinates=None):
        """
        Initialize a Molecule object.

        Raises
        ------
        AssertionError
            If the length of the atoms and coordinates lists do not match.
        """

        self.atoms = atoms or []
        if len(self.atoms) == 1:
            self.is_atom = True
        else:
            self.is_atom = False

        self.coordinates = coordinates or []
        assert len(self.atoms) == len(self.coordinates)

        self.atomic_info = "null"
        if self.is_atom:
            self.atomic_info = Element(self.atoms[0])

    def as_dict(self) -> dict:
        """
        Returns a dictionary representation of the molecule.

        Returns
        -------
        dict
            A dictionary with the following keys:
            - atoms: list of atomic symbols
            - coordinates: list of tuples representing the cartesian coordinates of the atoms in the molecule
            - is_atom: boolean indicating whether the molecule is a single atom
            - atomic_info: dictionary representing the atomic properties of the atom in the molecule, if the molecule is a single atom.
        """
        d = {
            "atoms": self.atoms,
            "coordinates": self.coordinates,
            "is_atom": self.is_atom,
            "atomic_info": self.atomic_info.as_dict() if self.is_atom else "null",
        }
        return d



import os

import yaml

__author__ = "YingXing Cheng"
__email__ = "yingxing.cheng@ugent.be"
__maintainer__ = "YingXing Cheng"
__maintainer_email__ = "yingxing.cheng@ugent.be"
__version__ = "1.0.0"

SETTINGS_FILE = os.path.join(os.path.expanduser("~"), ".pmgrc.yaml")


def _load_pmg_settings():
    try:
        with open(SETTINGS_FILE) as f:
            d = yaml.safe_load(f)
    except OSError:
        # If there are any errors, default to using environment variables
        # if present.
        d = {}
        # for k, v in os.environ.items():
        #     if k.startswith("PMG_"):
        #         d[k] = v
        #     elif k in ["VASP_PSP_DIR", "MAPI_KEY", "DEFAULT_FUNCTIONAL"]:
        #         d["PMG_" + k] = v
    d = d or {}
    return dict(d)


SETTINGS = _load_pmg_settings()

from pydirac.analysis import *
from pydirac.cli import *
from pydirac.core import *
from pydirac.io import *



from pydirac.analysis.polarizability import get_polarizability

__all__ = ["get_atomDB"]


def get_atomDB(args):
    if isinstance(args.dir_list, str):
        dirname_lis = [args.dir_list]
    else:
        dirname_lis = args.dir_list

    for d in dirname_lis:
        dir_fullname = os.path.abspath(d)
        get_polarizability(dir_fullname, calc_dir_patters=args.patterns, deepth=args.deepth)



import re

from pydirac.core.periodic import Element

__all__ = ["get_keyword", "get_orbital_info", "get_energy", "get_symbol_and_charge"]


def get_keyword(element: str, task_type: str, orbital_info: str, delimer: str = "@") -> str:
    """
    Constructs a keyword for a given element, task type, and orbital information.

    Parameters:
    -----------
    element : str
        The element symbol for which the keyword is to be constructed.
    task_type : str
        The type of task associated with the keyword.
    orbital_info : str
        The orbital information to be included in the keyword.
    delimer : str, optional
        The delimiter to be used to join the different components of the keyword.

    Returns:
    --------
    keyword : str
        The constructed keyword.

    Example:
    --------
    >>> get_keyword('C', 'energy', '(core 2)[vir 3]')
    'C@energy@(core 2)[vir 3]'
    """
    return delimer.join([Element(element).symbol, task_type, orbital_info])


def get_orbital_info(nocc: int, nvir: int) -> str:
    """
    Constructs the orbital information string for a given number of occupied and virtual orbitals.

    Parameters:
    -----------
    nocc : int
        The number of occupied orbitals.
    nvir : int
        The number of virtual orbitals.

    Returns:
    --------
    orbital_info : str
        The constructed orbital information string.

    Example:
    --------
    >>> get_orbital_info(2, 3)
    '(core 2)[vir 3]'
    """
    return f"(core {nocc})[vir {nvir}]"


def get_symbol_and_charge(filename="atom.mol"):
    """
    Get atom symbols and charges from a `.mol` file.

    Parameters:
    -----------
    filename : str, optional
        The name of the `.mol` file to read.

    Returns:
    --------
    atoms : list of tuple of str and float
        A list of tuples containing the symbol and charge of each atom in the file.

    Example:
    --------
    >>> get_symbol_and_charge('atom.mol')
    [('C', 6.0), ('H', 1.0), ('H', 1.0), ('H', 1.0)]
    """

    with open(filename) as f:
        context = f.read()

    # Regular expression to match the symbol and charge of each atom in the file
    pattern = r"^\s+(\d+)\.\s+(\d+\.?\d?)\s+"
    re_obj = re.compile(pattern)
    atoms = re.findall(pattern, context, re.MULTILINE)

    # Convert charge from string to float and return a list of (symbol, charge) tuples
    atoms = [(re.sub(r"[0-9]+", "", symbol), float(charge)) for symbol, charge in atoms]

    return atoms


def get_energy(filename, method="CCSD(T)"):
    """
    Get the energy from a quantum chemical calculation output file.

    Given a quantum chemical calculation output file specified by `filename` and
    the type of energy to retrieve given by `method`, the options are "CCSD(T)",
    "CCSD", "MP2", or "SCF".

    Parameters:
    -----------
    filename : str
        The name of the quantum chemical calculation output file.
    method : str, optional
        The type of energy to retrieve. Default is "CCSD(T)".

    Returns:
    --------
    energy : float
        The energy value.

    Raises:
    -------
    RuntimeError :
        If the energy for the specified method is not found in the file.

    Example:
    --------
    >>> get_energy('calc.out', method='CCSD(T)')
    -245.186792800
    """

    with open(filename) as f:
        context = f.read()

    # Regular expression to match the energy for the specified method
    if "(" and ")" in method:
        # Escape parentheses
        method = method.replace("(", r"\(").replace(")", r"\)")
    pattern = r"^@.*{method} energy[\s:]*(-?[\d\.]+)".format(**{"method": method})
    energy = re.findall(pattern, context, re.MULTILINE)

    if energy:
        return float(energy[-1])
    else:
        raise RuntimeError(f"Did not find energy for {method} in {filename}")



import abc
import glob
import os
import shutil
import warnings
from copy import deepcopy
from pathlib import Path

import importlib_resources
from monty.json import MSONable
from monty.serialization import loadfn

from pydirac.core.molecule import Molecule
from pydirac.core.settings import Settings
from pydirac.io.inputs import DiracInput, Inp, Mol
from pydirac.io.outputs import Output

__all__ = ["DiracInput", "AtomicDHFSet", "AtomicCCSet", "AtomicCISet"]

dir_path = os.path.dirname(__file__)

MODULE_DIR = Path(__file__).resolve().parent

"""
This module defines the DiracInputSet abstract base class and a concrete
implementation for the parameters developed and tested by the author.
The basic concept behind an input set is to specify a scheme to generate
a consistent set of DIRAC inputs from a structure without further user
intervention. This ensures comparability across runs.
"""


class DiracInputSet(MSONable, metaclass=abc.ABCMeta):
    """
    Abstract base class representing a set of Dirac input parameters with a
    structure supplied as initialization parameters.

    Attributes:
        inp (property): Inp object.
        mol (property): Mol object.
    """

    @property
    @abc.abstractmethod
    def inp(self):
        """Inp object"""
        pass

    @property
    @abc.abstractmethod
    def mol(self):
        """Mol object
        Returns:
        """
        pass

    def get_dirac_input(self):
        """Return a :class:`~pydirac.inputs.DiracInput` instance from the input and
        molecule objects.

        Returns:
            DiracInput: DiracInput instance.
        """
        return DiracInput(inp=self.inp, mol=self.mol)

    def write_input(self, output_dir, make_dir_if_not_present=True):
        """
        Write a set of DIRAC input to a directory.

        Args:
            output_dir (str): Path to the output directory.
            make_dir_if_not_present (bool): If True, create the directory if not present.

        Returns:
            None
        """
        dinput = self.get_dirac_input()
        dinput.write_input(output_dir, make_dir_if_not_present=make_dir_if_not_present)


def _load_yaml_config(fname):
    """
    Load a YAML configuration file and merge it with its parent configurations.

    Parameters
    ----------
    fname : str
        The name of the YAML configuration file to load.

    Returns
    -------
    dict
        The merged YAML configuration dictionary.

    """
    config = loadfn(str(importlib_resources.files("pydirac.data.template") / f"{fname}.yaml"))
    if "PARENT" in config:
        parent_config = _load_yaml_config(config["PARENT"])
        for k, v in parent_config.items():
            if k not in config:
                config[k] = v
            elif isinstance(v, dict):
                v_new = config.get(k, {})
                v_new.update(v)
                config[k] = v_new
    return config


class DictSet(DiracInputSet):
    """
    Concrete implementation of :class:`~pydirac.input_sets.DiracInputSet` that generates
    DIRAC input sets from a dictionary of settings and a molecule.

    Parameters
    ----------
    molecule : pymatgen.Molecule
        The molecule object.
    config_dict : dict
        A dictionary of input and molecule settings.
    files_to_transfer : dict, optional
        A dictionary of files to transfer.
    user_inp_settings : dict, optional
        A dictionary of user-defined input settings.
    user_mol_settings : dict, optional
        A dictionary of user-defined molecule settings.
    use_structure_charge : bool, optional
        Whether to use the structure charge.

    Attributes
    ----------
    inp : Inp
        Inp object.
    molecule : pymatgen.Molecule
        The molecule object.
    mol : Mol
        Mol object.

    Methods
    -------
    write_input(output_dir, make_dir_if_not_present=True)
        Write a set of DIRAC input to a directory.

    """

    def __init__(
        self,
        molecule,
        config_dict,
        files_to_transfer=None,
        user_inp_settings=None,
        user_mol_settings=None,
        use_structure_charge=False,
    ):
        """
        Constructor for :class:`~pydirac.input_sets.DictSet`.

        Parameters
        ----------
        molecule : pymatgen.Molecule
            The molecule object.
        config_dict : dict
            A dictionary of input and molecule settings.
        files_to_transfer : dict, optional
            A dictionary of files to transfer.
        user_inp_settings : dict, optional
            A dictionary of user-defined input settings.
        user_mol_settings : dict, optional
            A dictionary of user-defined molecule settings.
        use_structure_charge : bool, optional
            Whether to use the structure charge.

        """
        self._molecule = molecule
        self._config_dict = deepcopy(config_dict)
        self.files_to_transfer = files_to_transfer or {}
        self.user_inp_settings = user_inp_settings or {}
        self.user_mol_settings = user_mol_settings or {}
        self.use_molecule_charge = use_structure_charge

    @property
    def inp(self) -> Inp:
        """
        Inp object.

        Returns
        -------
        Inp
            Inp instance.

        """
        settings = dict(self._config_dict["inp"])
        inp = Inp(settings)
        return inp

    @property
    def molecule(self) -> Molecule:
        return self._molecule

    @property
    def mol(self) -> Mol:
        """
        Mol object.

        Returns
        -------
        Mol
            Mol instance.

        """
        mol = Mol(self.molecule)
        mol_settings = dict(self._config_dict["mol"])
        mol.basis_type = mol_settings.get("basis_type", "dyall.v2z")

        if "basis_type" in self.user_mol_settings:
            mol.basis_type = self.user_mol_settings["basis_type"]
        if "molecule" in self.user_mol_settings:
            mol.basis_type = self.user_mol_settings["molecule"]
        if "basis_lib" in self.user_mol_settings:
            mol.basis_lib = self.user_mol_settings["basis_lib"]
        if "comment" in self.user_mol_settings:
            mol.comment = self.user_mol_settings["comment"]
        return mol

    def __str__(self):
        """
        Return the class name.

        Returns
        -------
        str
            The class name.

        """
        return self.__class__.__name__

    def write_input(self, output_dir, make_dir_if_not_present=True):
        """
        Write a set of DIRAC input to a directory.

        Parameters
        ----------
        output_dir : str
            The path of the directory to write the input files to.
        make_dir_if_not_present : bool, optional
            Whether to create the directory if it does not exist.

        Returns
        -------
        None

        """
        super().write_input(output_dir=output_dir, make_dir_if_not_present=make_dir_if_not_present)
        for k, v in self.files_to_transfer.items():
            shutil.copy2(v, str(Path(output_dir) / k))


class AtomicDHFSet(DictSet):
    """
    A class representing a set of input parameters for an atomic Dirac-Hartree-Fock (DHF) calculation.

    Parameters
    ----------
    molecule : Molecule
        The molecule to compute the DHF calculation for.
    hamiltonian_mode : str, optional
        The type of Hamiltonian to use. Must be one of "2C" (two-component) or "4C" (four-component).
        Defaults to "4C".
    is_spinfree : bool, optional
        Whether to use a spin-free Hamiltonian. Defaults to False.
    is_ff : bool, optional
        Whether to include a frozen-core approximation. Defaults to False.
    ff_mode : str, optional
        The type of frozen-core approximation to use. Must be one of "D" (dipole approximation) or "Q"
        (quadrupole approximation). Defaults to "D".
    zff : float, optional
        The factor by which to scale the frozen-core potential. Only used if is_ff is True. Defaults to None.
    **kwargs
        Additional keyword arguments to be passed to the superclass constructor.

    Attributes
    ----------
    SUPPORTED_FF_MODES : tuple of str
        The supported frozen-core approximation modes.
    SUPPORTED_HAMILTONIAN_MODES : tuple of str
        The supported Hamiltonian modes.

    Raises
    ------
    ValueError
        If the ff_mode parameter is not one of the supported modes.

    Notes
    -----
    This class inherits from DictSet and uses the configuration file "AtomicDHFSet" to populate its input dictionary.

    """

    CONFIG = _load_yaml_config("AtomicDHFSet")

    SUPPORTED_FF_MODES = ("D", "Q")
    SUPPORTED_HAMILTONIAN_MODES = ("2C", "4C")

    def __init__(
        self,
        molecule,
        hamiltonian_mode="4C",
        is_spinfree=False,
        is_ff=False,
        ff_mode="D",
        zff=None,
        **kwargs,
    ):
        super().__init__(molecule, AtomicDHFSet.CONFIG, **kwargs)
        self.hamiltonian_mode = hamiltonian_mode
        self.is_spinfree = is_spinfree
        self.is_ff = is_ff
        self.ff_mode = ff_mode
        self.zff = zff
        self.kwargs = kwargs

        if self.ff_mode not in AtomicDHFSet.SUPPORTED_FF_MODES:
            raise ValueError(
                "{} not one of the support modes : {}".format(
                    self.ff_mode, AtomicDHFSet.SUPPORTED_FF_MODES
                )
            )

    @property
    def inp(self):
        """
        The input dictionary for the DHF calculation.

        Returns
        -------
        dict
            The input dictionary.
        """
        parent_inp = super().inp

        # update open shell setup
        if self.molecule.is_atom:
            elec_dict = self.molecule.atomic_info.get_dirac_shell_str(return_dict=True)
            for k in ["CLOSED SHELL", "OPEN SHELL"]:
                if k in elec_dict:
                    parent_inp[parent_inp.wf_tag]["SCF"].update({k: elec_dict[k]})
        else:
            # default is closed shell
            pass

        # update hamiltonian, 4c and spinfree
        parent_inp.update(
            {
                "HAMILTONIAN": {
                    "DOSSSS": self.hamiltonian_mode == "4C",
                    "X2C": self.hamiltonian_mode == "2C",
                    "NOSPIN": self.is_spinfree,
                }
            }
        )

        # update operator
        if self.is_ff:
            if self.zff is None:
                zff = "zff"
            else:
                zff = float(self.zff)
            if "HAMILTONIAN" in parent_inp:
                if self.ff_mode.upper() == "Q":
                    operator_dict = {
                        "OPERATOR": [
                            " Theta quadru-field",
                            " DIAGONAL",
                            " ZZTHETA",
                            " COMFACTOR",
                            f" {zff}",
                        ]
                    }
                else:
                    operator_dict = {"OPERATOR": [" ZDIPLEN", " COMFACTOR", f" {zff}"]}
                parent_inp["HAMILTONIAN"].update(operator_dict)
        return parent_inp


class AtomicCCSet(AtomicDHFSet):
    """
    Specification of reference determinant, type of calculation, and general settings.

    Default is energy calculation.

    Parameters
    ----------
    molecule : AtomicStruct or Molecule
        The atomic or molecular structure.
    e_min : float, optional
        Minimum energy of the active space. Default is -20.0.
    e_max : float, optional
        Maximum energy of the active space. Default is 25.0.
    e_error : float, optional
        Convergence threshold for energy. Default is 0.01.
    maxit : int, optional
        Maximum number of iterations. Default is 60.
    ntol : int, optional
        Convergence threshold for the norm of the residual. Default is 10.
    nelec : tuple of int, optional
        Number of electrons for closed-shell systems. Default is None.
    nelec_open : tuple of int, optional
        Number of electrons for open-shell systems. Default is None.
    nel_f1 : tuple of int, optional
        Number of frozen core electrons in the first fragment. Default is None.
    nel_f2 : tuple of int, optional
        Number of frozen core electrons in the second fragment. Default is None.
    no_T : bool, optional
        If True, does not include the T amplitudes. Default is None.
    print_level : int, optional
        Level of detail printed. Default is 1.
    prev_inp : str or Inp, optional
        Path to previous input file or the previous input file object. Default is None.
    prev_mol : str or Mol, optional
        Path to previous molecular file or the previous molecular file object. Default is None.
    **kwargs
        Additional arguments to pass to AtomicDHFSet.

    Attributes
    ----------
    e_min : float
        Minimum energy of the active space.
    e_max : float
        Maximum energy of the active space.
    e_error : float
        Convergence threshold for energy.
    maxit : int
        Maximum number of iterations.
    ntol : int
        Convergence threshold for the norm of the residual.
    nelec : tuple of int or None
        Number of electrons for closed-shell systems.
    nelec_open : tuple of int or None
        Number of electrons for open-shell systems.
    nel_f1 : tuple of int or None
        Number of frozen core electrons in the first fragment.
    nel_f2 : tuple of int or None
        Number of frozen core electrons in the second fragment.
    no_T : bool or None
        If True, does not include the T amplitudes.
    print_level : int
        Level of detail printed.
    prev_inp : Inp or None
        Previous input file object or None.
    prev_mol : Mol or None
        Previous molecular file object or None.
    kwargs : dict
        Additional arguments passed to AtomicDHFSet.

    """

    def __init__(
        self,
        molecule,
        e_min=-20.0,
        e_max=25.0,
        e_error=0.01,
        maxit=60,
        ntol=10,
        nelec=None,
        nelec_open=None,
        nel_f1=None,
        nel_f2=None,
        no_T=None,
        print_level=1,
        prev_inp=None,
        prev_mol=None,
        **kwargs,
    ):
        super().__init__(molecule, **kwargs)
        # self._config_dict['inp'].update()
        if isinstance(prev_inp, str):
            prev_inp = Inp.from_file(prev_inp)
        if isinstance(prev_mol, str):
            prev_mol = Mol.from_file(prev_mol)

        self.e_min = e_min
        self.e_max = e_max
        self.e_error = e_error
        self.maxit = maxit
        self.ntol = ntol
        self.nelec = nelec
        self.nelec_open = nelec_open
        self.nel_f1 = nel_f1
        self.nel_f2 = nel_f2
        self.no_T = no_T
        self.print_level = print_level
        self.prev_inp = prev_inp
        self.prev_mol = prev_mol
        self.kwargs = kwargs

    @property
    def inp(self):
        """
        Generates an input file object with the settings required for a CC calculation.

        Returns
        -------
        Inp
            Input file object.
        """
        parent_inp = super().inp
        inp = self.prev_inp if self.prev_inp is not None else parent_inp

        s = Settings(inp)
        # no matter whether previous inp has 4index
        s.DIRAC["4INDEX"] = True
        s[inp.wf_tag].RELCCSD = True
        # s.integrals.readinp.uncontract = True
        s.INTEGRALS.READINP = "UNCONTRACT"
        if "active" not in s.MOLTRA:
            s.MOLTRA.ACTIVE = f"{self.e_min} {self.e_max} {self.e_error}"

        # RELCC setup
        RELCC = Settings()
        RELCC.ENERGY = True
        RELCC.PRINT = self.print_level
        RELCC.CCENER = True
        RELCC.MAXIT = self.maxit
        RELCC.NTOL = self.ntol
        if "NOSDT" not in s.RELCC and self.no_T is True:
            RELCC.NOSDT = self.no_T

        nelec = self.nelec_open or self.nelec
        if nelec and len(nelec):
            RELCC.NELEC = " ".join([str(i) for i in nelec])
        if self.nel_f1 and len(self.nel_f1):
            RELCC.NEL_F1 = " ".join([str(i) for i in self.nel_f1])
        if self.nel_f2 and len(self.nel_f2):
            RELCC.NEL_F2 = " ".join([str(i) for i in self.nel_f1])
        s.RELCC = RELCC
        # update RELCC setup
        inp.update(s.as_dict())
        inp.update(self.user_inp_settings)
        return inp

    @property
    def mol(self):
        """
        Returns the molecular file object.

        Returns
        -------
        Mol
            Molecular file object.
        """
        parent_mol = super().mol
        this_mol = self.prev_mol if self.prev_mol is not None else parent_mol
        this_mol.update(self.user_mol_settings)
        return this_mol

    def override_from_prev_calc(self, prev_calc_dir=".", no_scf=False):
        """
        Update the input set to include settings from a previous calculation.

        Parameters
        ----------
        prev_calc_dir : str, optional
            Path to the previous calculation directory. Default is ".".
        no_scf : bool, optional
            If True, does not perform SCF calculations. Default is False.

        Returns
        -------
        AtomicCCSet
            Input set with updated settings.
        """
        output = get_diracrun_output(prev_calc_dir)

        self.prev_inp = output.inp
        self.prev_mol = output.mol
        # TODO: we do not need to override a DHF calculation, only one maybe .SCF
        if no_scf:
            self.prev_inp[self.prev_inp.wf_tag]["SCF"].update({"_en": False})

        # TODO, the only need to change is the number of active electrons
        # specifiec by .NELEC keyword or NEL_F1
        return self

    @classmethod
    def from_prev_calc(cls, prev_calc_dir, ff_mode="D", no_scf=False, **kwargs):
        """
        Create an AtomicCCSet object using settings from a previous calculation.

        Parameters
        ----------
        prev_calc_dir : str
            Path to the previous calculation directory.
        ff_mode : str, optional
            Fragmentation mode. Default is "D".
        no_scf : bool, optional
            If True, does not perform SCF calculations. Default is False.
        **kwargs
            Additional arguments to pass to AtomicDHFSet.

        Returns
        -------
        AtomicCCSet
            Input set with updated settings from previous calculation.
        """
        input_set = cls(_dummy_structure, ff_mode=ff_mode, **kwargs)
        return input_set.override_from_prev_calc(prev_calc_dir, no_scf)


class AtomicCISet(AtomicDHFSet):
    """
    Specifies the reference determinant and type of calculation for CI calculations.

    Inherits from AtomicDHFSet.

    Parameters
    ----------
    molecule : str or pymatgen.Molecule
        A string or pymatgen.Molecule object representing the molecule to calculate.
    prev_inp : str or Inp, optional
        A string or Inp object representing the previous input file to be used as a template.
        Default is None.
    prev_mol : str or Mol, optional
        A string or Mol object representing the previous molecular file to be used as a template.
        Default is None.
    **kwargs
        Additional keyword arguments to pass to AtomicDHFSet.

    Attributes
    ----------
    prev_inp : Inp
        The previous input file to be used as a template.
    prev_mol : Mol
        The previous molecular file to be used as a template.
    kwargs : dict
        Additional keyword arguments passed to AtomicDHFSet.

    Notes
    -----
    The default calculation is energy calculation.
    """

    def __init__(self, molecule, prev_inp=None, prev_mol=None, **kwargs):
        super().__init__(molecule, **kwargs)
        if isinstance(prev_inp, str):
            prev_inp = Inp.from_file(prev_inp)
        if isinstance(prev_mol, str):
            prev_mol = Mol.from_string(prev_mol)

        self.prev_inp = prev_inp
        self.prev_mol = prev_mol
        self.kwargs = kwargs

    @property
    def inp(self):
        """
        Generates an input file object with the settings required for a CI calculation.

        Returns
        -------
        Inp
            Input file object.
        """
        inp = super().inp

        # Step 1: the default KRCI setup
        s = Settings(inp)
        # s.dirac.analyze = True
        # wf_tag = inp.wf_tag or 'WAVE FUNCTIONS'
        # s.dirac[wf_tag] = True
        # s.analyze.mulpop._en = True
        # s.analyze.mulpop.vecpop = '1..oo'
        # s.integrals.readinp = 'uncontract'

        # wave_func = Settings()
        # wave_func['KR CI'] = True
        # wave_func.resolve = True
        # s[wf_tag] = wave_func

        # krci = Settings()
        # krci['CI PROGRAM'] = 'LUCIAREL'
        # # TODO: for different elements, the krci setup is different
        # # krci.inactive = 1
        # # krci['GAS SHELLS'] = [3, '0 2 / 1', '1 3 / 3', '3 3 / 10']
        # krci['MAX CI'] = 60
        # krci['NOOCCN'] = True
        # krci.nooccn = True
        # krci.rstrci = 0
        # # TODO: for different element, the number of roots is also different
        # krci['CIROOTS_id_0'] = '3  3'
        # krci['CIROOTS_id_1'] = '4  3'
        # s[wf_tag].krcicalc = krci
        inp.update(s.as_dict())

        # Step 2: if the previous inp exist, we use the previous one
        if self.prev_inp:
            # pre_s = Settings(self.prev_inp)
            # print(pre_s)
            # inp.update(pre_s.as_dict())
            inp = self.prev_inp
        else:
            warnings.warn("For KRCI module, more detail need to be considered")
            return None

        # Step 3: user setting
        inp.update(self.user_inp_settings)
        return inp

    @property
    def mol(self):
        """
        Returns the molecular file object.

        Returns
        -------
        Mol
            Molecular file object.
        """
        parent_mol = super().mol
        this_mol = self.prev_mol if self.prev_mol is not None else parent_mol
        this_mol.update(self.user_mol_settings)
        return this_mol

    def override_from_prev_calc(self, prev_calc_dir=".", no_scf=False):
        """
        Update the input set to include settings from a previous calculation.

        Parameters
        ----------
        prev_calc_dir : str, optional
            Path to the previous calculation directory. Default is ".".
        no_scf : bool, optional
            If True, does not perform SCF calculations. Default is False.

        Returns
        -------
        AtomicCISet
            Input set with updated settings.
        """
        output = get_diracrun_output(prev_calc_dir)

        self.prev_inp = output.inp
        self.prev_mol = output.mol
        if no_scf:
            self.prev_inp[self.prev_inp.wf_tag]["SCF"].update({"_en": False})
        return self

    @classmethod
    def from_prev_calc(cls, prev_calc_dir, ff_mode="D", no_scf=False, **kwargs):
        """
        Create an AtomicCISet object using settings from a previous calculation.

        Parameters
        ----------
        prev_calc_dir : str
            Path to the previous calculation directory.
        ff_mode : str, optional
            Fragmentation mode. Default is "D".
        no_scf : bool, optional
            If True, does not perform SCF calculations. Default is False.
        **kwargs
            Additional arguments to pass to AtomicDHFSet.

        Returns
        -------
        AtomicCISet
            Input set with updated settings from previous calculation.
        """
        input_set = cls(_dummy_structure, ff_mode=ff_mode, **kwargs)
        return input_set.override_from_prev_calc(prev_calc_dir, no_scf)

    @staticmethod
    def get_info_for_s_block(out: Output, e_min, e_max, calc_type="quadrupole"):
        """
        Get information for the s block.

        Parameters
        ----------
        out : Output
            The output file to parse.
        e_min : float
            The minimum energy.
        e_max : float
            The maximum energy.
        calc_type : str, optional
            The type of calculation. Default is "quadrupole".

        Returns
        -------
        nb_active_elec: int
            The number of electrons in the GAS.
        gas_list: list
            A list for GAS setup.
        root_list: list
            Root for different symmetries.
        """
        atomic_info = out.mol.molecule.atomic_info
        mos = out.mos
        assert atomic_info.block == "s"

        (
            nb_closed_elec,
            nb_open_elec,
            nb_total_elec,
            nb_closed_shell,
            nb_open_shell,
            nb_vir_shell,
        ) = mos.get_ao_and_elec(e_min, e_max)

        assert nb_closed_elec == nb_closed_shell
        assert nb_open_elec <= nb_open_shell
        period = atomic_info.period

        gas_list = []
        root_list = []

        if nb_open_shell == 0:
            # for nobel gas elements
            if period == 1:
                # He
                nb_active_elec = 2
                gas1 = "0 2 / 1"
                gas2 = f"2 2 / {nb_vir_shell // 2}"
                gas_list.extend([gas1, gas2])
            elif period == 2:
                # Be
                nb_active_elec = 4
                gas1 = "2 4 / 2"
                gas2 = f"4 4 / {nb_vir_shell // 2}"
                gas_list.extend([gas1, gas2])
            elif period == 3:
                # Mg
                nb_active_elec = 10
                gas1 = "0 2 / 1"
                gas2 = "8 10 / 4"
                gas3 = f"10 10 / {nb_vir_shell // 2}"
                gas_list.extend([gas1, gas2, gas3])
                pass
            else:
                # Ca Sr Ba Ra
                # including d electrons
                nb_active_elec = 20
                gas1 = "10 12 / 6 ! (n-1)s2 (n-2)d10"
                gas2 = "18 20 / 4 ! (n-1)p6 (n)s2"
                gas3 = f"20 20 / {nb_vir_shell // 2} ! virtual orbitals"
                gas_list.extend([gas1, gas2, gas3])
        else:
            # for s-block open-shell elements
            if period > 3:
                # K Rb Cs Fr
                # including d electrons
                nb_active_elec = 19
                gas1 = "10 12 / 6 ! (n-1)s2 (n-2)d10"
                gas2 = "17 19 / 4 ! (n-1)p6 (n)s1"
                gas3 = f"19 19 / {nb_vir_shell // 2} ! virtual orbitals"
                gas_list.extend([gas1, gas2, gas3])
            elif period == 3:
                # for Na
                nb_active_elec = 9
                gas1 = "0 2 / 1"
                gas2 = "7 9 / 4"
                gas3 = f"9 9 / {nb_vir_shell // 2}"
                gas_list.extend([gas1, gas2, gas3])
            elif period == 2:
                # for Li
                nb_active_elec = 3
                gas1 = "1 3 / 2 ! 1s"
                gas2 = f"3 3 / {nb_vir_shell // 2}"
                gas_list.extend([gas1, gas2])
            else:
                # H
                assert period == 1
                nb_active_elec = 1
                gas1 = "0 1 / 1 ! 1s"
                gas2 = f"1 1 / {nb_vir_shell // 2}"
                gas_list.extend([gas1, gas2])

        # specify many roots should we compute
        if nb_open_elec == 1:
            # IA: H Li Na K Rb Cs Fr
            ciroot1 = "3 1"
            root_list.extend([ciroot1])
        elif nb_open_elec == 0:
            # IIA: He Be Mg Ca Sr Ba Ra
            ciroot1 = "1 1"
            # ciroot2 = '2 1'
            root_list.extend([ciroot1])
        else:
            raise RuntimeError(
                "The number of electrons in open shell " "is wrong: {}".format(nb_open_elec)
            )

        return nb_active_elec, gas_list, root_list

    @staticmethod
    def get_info_for_p_block(out: Output, e_min, e_max, calc_type="quadrupole"):
        """
        Get information for the p block.

        Parameters
        ----------
        out : Output
            The output file to parse.
        e_min : float
            The minimum energy.
        e_max : float
            The maximum energy.
        calc_type : str, optional
            The type of calculation. Default is "quadrupole".

        Returns
        -------
        nb_active_elec: int
            The number of electrons in the GAS.
        gas_list: list
            A list for GAS setup.
        root_list: list
            Root for different symmetries.
        """
        atomic_info = out.mol.molecule.atomic_info
        mos = out.mos
        assert atomic_info.block == "p"

        (
            nb_closed_elec,
            nb_open_elec,
            nb_total_elec,
            nb_closed_shell,
            nb_open_shell,
            nb_vir_shell,
        ) = mos.get_ao_and_elec(e_min, e_max)

        assert nb_closed_elec == nb_closed_shell
        assert nb_open_elec <= nb_open_shell
        period = atomic_info.period

        gas_list = []
        root_list = []

        if nb_open_shell == 0:
            # for nobel gas elements
            # if period == 1:
            #     # He
            #     nb_active_elec = 2
            #     gas1 = '0 2 / 1'
            #     gas2 = '2 2 / {0}'.format(nb_vir_shell // 2)
            #     gas_list.extend([gas1, gas2])
            if period == 2:
                # Ne
                nb_active_elec = 10
                gas1 = "2 4 / 2"
                gas2 = "8 10 / 3"
                gas3 = f"10 10 / {nb_vir_shell // 2}"
                gas_list.extend([gas1, gas2, gas3])
            else:
                # Ar, Kr, Xe, Rn, Og
                # including d electrons
                nb_active_elec = 18
                gas1 = "10 12 / 6 ! (n-1)s2 (n-2)d10"
                gas2 = "16 18 / 3 ! (n-1)p6"
                gas3 = f"18 18 / {nb_vir_shell // 2} ! virtual orbitals"
                gas_list.extend([gas1, gas2, gas3])
        else:
            # for p-block open-shell elements
            if period > 3:
                # including d electrons
                nb_active_elec = 10 + 2 + nb_open_elec
                gas1 = "10 12 / 6 ! (n-1)s2 (n-2)d10"
                gas2 = f"{nb_open_elec + 10} {nb_open_elec + 12} / 3 ! (n-1)p6"
                gas3 = "{} {} / {} ! virtual orbitals".format(
                    nb_open_elec + 12, nb_open_elec + 12, nb_vir_shell // 2
                )
                gas_list.extend([gas1, gas2, gas3])
            elif period == 3:
                # for Al Si P S Cl
                nb_active_elec = 2 + nb_open_elec
                gas1 = "0 2 / 1"
                gas2 = f"{nb_open_elec} {nb_open_elec + 2} / 3"
                gas3 = f"{nb_open_elec + 2} {nb_open_elec + 2} / {nb_vir_shell // 2}"
                gas_list.extend([gas1, gas2, gas3])
            else:
                # without d electrons
                assert period == 2
                # for B C N O F
                nb_active_elec = nb_total_elec
                gas1 = "2 4 / 2 ! 1s2s"
                gas2 = f"{nb_total_elec - 2} {nb_total_elec} / 3 ! 2p"
                gas3 = f"{nb_total_elec} {nb_total_elec} / {nb_vir_shell // 2}"
                gas_list.extend([gas1, gas2, gas3])

        # specify many roots should we compute
        if nb_open_elec == 1:
            # IIIA: B Al Ga In Tl Nh
            ciroot1 = "3 3"
            ciroot2 = "4 3"
            root_list.extend([ciroot1, ciroot2])
        elif nb_open_elec == 2:
            # IVA: C Si Ge Sn Pb Fl
            if calc_type == "dipole":
                # DIRAC bugs, only one root does not get exact gs state
                ciroot1 = "1 2"
                # ciroot2 = '2 1'
                root_list.extend([ciroot1])
            elif calc_type == "quadrupole":
                # ciroot1 = '1 9'
                # ciroot2 = '2 6'
                ciroot1 = "1 2"
                ciroot2 = "2 2"
                root_list.extend([ciroot1, ciroot2])
            else:
                raise TypeError('Calc_type wrong! "dipole" or "quadrupole"')
        elif nb_open_elec == 3:
            # VA: N P As Sb Bi Mc
            if calc_type == "dipole":
                ciroot1 = "3 2"
                ciroot2 = "4 2"
                root_list.extend([ciroot1, ciroot2])
            elif calc_type == "quadrupole":
                # ciroot1 = '3 10'
                # # ciroot2 = '4 10'
                ciroot1 = "3 2"
                # ciroot2 = '4 2'
                root_list.extend([ciroot1])
            else:
                raise TypeError('Calc_type wrong! "dipole" or "quadrupole"')
        elif nb_open_elec == 4:
            # # VIA: O S Se Te Po Lv
            if calc_type == "dipole":
                ciroot1 = "1 3"
                ciroot2 = "2 2"
                root_list.extend([ciroot1, ciroot2])
            elif calc_type == "quadrupole":
                # ciroot1 = '1 9'
                # ciroot2 = '2 6'
                ciroot1 = "1 3"
                ciroot2 = "2 2"
                root_list.extend([ciroot1, ciroot2])
            else:
                raise TypeError('Calc_type wrong! "dipole" or "quadrupole"')
        elif nb_open_elec == 5:
            # VIIA: F Cl Br I At Ts
            ciroot1 = "3 3"
            ciroot2 = "4 3"
            root_list.extend([ciroot1, ciroot2])
        elif nb_open_elec == 0:
            # closed-shell elementss
            ciroot1 = "1 1"
            # ciroot2 = '2 1'
            root_list.extend([ciroot1])
        else:
            raise RuntimeError(
                "The number of electrons in open shell " "is wrong: {}".format(nb_open_elec)
            )

        return nb_active_elec, gas_list, root_list

    @staticmethod
    def get_info_for_d_block(out: Output, e_min, e_max, calc_type="quadrupole"):
        """
        Get information for the d block.

        Parameters
        ----------
        out : Output
            The output file to parse.
        e_min : float
            The minimum energy.
        e_max : float
            The maximum energy.
        calc_type : str, optional
            The type of calculation. Default is "quadrupole".

        Returns
        -------
        nb_active_elec: int
            The number of electrons in the GAS.
        gas_list: list
            A list for GAS setup.
        root_list: list
            Root for different symmetries.
        """
        atomic_info = out.mol.molecule.atomic_info
        mos = out.mos
        assert atomic_info.block == "d"

        (
            nb_closed_elec,
            nb_open_elec,
            nb_total_elec,
            nb_closed_shell,
            nb_open_shell,
            nb_vir_shell,
        ) = mos.get_ao_and_elec(e_min, e_max)

        assert nb_closed_elec == nb_closed_shell
        assert nb_open_elec <= nb_open_shell
        period = atomic_info.period

        gas_list = []
        root_list = []

        if nb_open_shell == 0:
            # IIB
            # Zn Cd Hg Cn
            # including d electrons
            nb_active_elec = 12
            gas1 = "8 10 / 5 ! (n-1)d10"
            gas2 = "10 12 / 1 ! (n)s2"
            gas3 = f"12 12 / {nb_vir_shell // 2} ! virtual orbitals"
            gas_list.extend([gas1, gas2, gas3])
        else:
            # for d-block open-shell elements
            if nb_open_shell == 2:
                # s is open
                # Cu Ag Au Rg
                assert nb_open_elec == 1
                nb_active_elec = 11
                gas1 = "8 10 / 5 ! (n-1)d10"
                gas2 = "9 11 / 1 ! (n)s1"
                gas3 = f"11 11 / {nb_vir_shell // 2} ! virtual orbitals"
                gas_list.extend([gas1, gas2, gas3])
            elif nb_open_shell == 10:
                # d is open
                nb_active_elec = 2 + nb_open_elec
                # normally there is no this case, since in DIRAC (n-1)d locates
                # before (n)s shells
                gas1 = "0 2 / 1 ! (n)s2"
                gas2 = f"{nb_open_elec} {nb_open_elec + 2} / 5 ! (n-1)d10"
                gas3 = "{0} {0} / {1} ! virtual orbitals".format(
                    nb_open_elec + 2, nb_vir_shell // 2
                )
                gas_list.extend([gas1, gas2, gas3])
            elif nb_open_shell == 12:
                # s + d are open
                nb_active_elec = nb_open_elec
                gas1 = f"{nb_open_elec - 2} {nb_open_elec} / 6 ! (n-1)d10 (n)s2"
                gas2 = "{0} {0} / {1} ! virtual orbitals".format(nb_open_elec, nb_vir_shell // 2)
                gas_list.extend([gas1, gas2])
            else:
                raise RuntimeError(f"Unknown number of open shell: {nb_open_shell}")

        # specify many roots should we compute
        if nb_open_elec == 1:
            if nb_open_shell == 2:
                # only s is open
                ciroot1 = "3 1"
                # ciroot2 = '2 1'
                root_list.extend([ciroot1])
            elif nb_open_shell == 12:
                raise NotImplementedError(
                    "For two open-shell d elements, " "there is no implement!"
                )
            else:
                raise RuntimeError("the number of open shell {} " "is error!".format(nb_open_shell))
        elif nb_open_elec == 0:
            # closed-shell elementss
            ciroot1 = "1 1"
            # ciroot2 = '2 1'
            root_list.extend([ciroot1])
        else:
            raise NotImplementedError(
                "For open-shell d elements and more than "
                "one open-shell element, it has not "
                "been implmented yet!"
            )

        return nb_active_elec, gas_list, root_list

    @staticmethod
    def get_info_for_f_block(out: Output, e_min, e_max, calc_type="quadrupole"):
        """
        Get information for the f block.

        Parameters
        ----------
        out : Output
            The output file to parse.
        e_min : float
            The minimum energy.
        e_max : float
            The maximum energy.
        calc_type : str, optional
            The type of calculation. Default is "quadrupole".

        Returns
        -------
        nb_active_elec: int
            The number of electrons in the GAS.
        gas_list: list
            A list for GAS setup.
        root_list: list
            Root for different symmetries.
        """
        atomic_info = out.mol.molecule.atomic_info
        mos = out.mos
        assert atomic_info.block == "f"

        (
            nb_closed_elec,
            nb_open_elec,
            nb_total_elec,
            nb_closed_shell,
            nb_open_shell,
            nb_vir_shell,
        ) = mos.get_ao_and_elec(e_min, e_max)

        assert nb_closed_elec == nb_closed_shell
        assert nb_open_elec <= nb_open_shell
        period = atomic_info.period

        gas_list = []
        root_list = []

        raise NotImplementedError("For open-shell f elements, it has not " "been implmented yet!")

    @classmethod
    def from_prev_dhf_calc(cls, filename_input, e_min=-10, e_max=10.0):
        """
        Create MRCI input based on the previous SCF calculation.

        Parameters
        ----------
        filename_input : str
            The name of the input file.
        e_min : float, optional
            The minimum energy. Default is -10.
        e_max : float, optional
            The maximum energy. Default is 10.0.

        Returns
        -------
        cls
            The MRCI input object.
        """

        out = Output(filename_input)
        out.parse_orbit()
        atom = out.mol.molecule
        # print(atom.is_atom)

        if not atom.is_atom:
            warnings.warn("This is not an atomic calculation!")
            return

        # this is for p block elements
        if atom.atomic_info.block == "p":
            nb_active_elec, gas_list, root_list = AtomicCISet.get_info_for_p_block(
                out, e_min=e_min, e_max=e_max, calc_type="quadrupole"
            )
        elif atom.atomic_info.block == "s":
            nb_active_elec, gas_list, root_list = AtomicCISet.get_info_for_s_block(
                out, e_min=e_min, e_max=e_max, calc_type="quadrupole"
            )
        elif atom.atomic_info.block == "d":
            nb_active_elec, gas_list, root_list = AtomicCISet.get_info_for_d_block(
                out, e_min=e_min, e_max=e_max, calc_type="quadrupole"
            )
        elif atom.atomic_info.block == "f":
            nb_active_elec, gas_list, root_list = AtomicCISet.get_info_for_f_block(
                out, e_min=e_min, e_max=e_max, calc_type="quadrupole"
            )
        else:
            raise RuntimeError("Unknown block element!")

        # print('total number electrons is {0}'.format(atom.atomic_info.Z))
        inactivate = (atom.atomic_info.Z - nb_active_elec) // 2
        nb_gas_shell = len(gas_list)
        nb_root_sym = len(root_list)

        krci_setup = {
            "CI PROGRAM": "LUCIAREL",
            "INACTIVE": str(inactivate),
            "GAS SHELLS": str(nb_gas_shell) + "\n" + "\n".join(gas_list),
            "MAX CI": 120,
            "MXCIVE": 60,
            "ANALYZ": True,
            "RSTRCI": "rstr",
            "CHECKP": True,
            "NOOCCN": True,
        }
        for i in range(nb_root_sym):
            krci_setup[f"CIROOTS_id_{i}"] = root_list[i]

        inp = out.inp
        if "HAMILTONIAN" in inp and "OPERATOR" in inp["HAMILTONIAN"]:
            inp["HAMILTONIAN"]["OPERATOR"][-1] = " zff"
        inp[inp.wf_tag]["KR CI"] = True
        inp[inp.wf_tag]["KRCICALC"] = krci_setup
        ci_obj = cls(atom)
        ci_obj.prev_inp = inp
        ci_obj.prev_mol = out.mol
        return ci_obj


_dummy_structure = Molecule([1], [[0.0, 0.0, 0.0]])


def get_diracrun_output(path) -> Output:
    """
    Get the output from a previous DIRAC calculation.

    Parameters
    ----------
    path : str
        The path to the previous calculation directory.

    Returns
    -------
    Output
        The output object from the previous calculation.
    """
    path = Path(path)
    outputs = list(glob.glob(str(path / "*.out")))
    if len(outputs) == 0:
        raise ValueError(f"Unable to get *.out from prev calculation in ${path}")
    outfile = sorted(outputs)[-1]
    return Output(outfile)


if __name__ == "__main__":
    import tempfile

    def main2():
        module_dir = os.path.dirname(__file__)
        output = str(Path(module_dir) / "../tests/data/Rn_q_so/d-aug-dyall.acv3z_+0.00001")

        # Use a temporary directory to store the output files
        with tempfile.TemporaryDirectory() as tmp_dir:
            old_cc = AtomicCCSet.from_prev_calc(
                output, no_scf=True, user_mol_settings={"basis_type": "s-aug-ANO-RCC"}
            )
            old_cc.write_input(output_dir=tmp_dir)
            new_cc = AtomicCCSet(
                molecule=Molecule([7], [[0.0, 0.0, 0.0]]),
                hamiltonian_mode="2C",
                is_spinfree=True,
                is_ff=True,
                ff_mode="D",
                user_mol_settings={"basis_type": "s-aug-ANO-RCC"},
            )
            new_cc.write_input(output_dir=tmp_dir)

    main()
    main2()



import os
import re
import subprocess
import warnings

from monty.json import MontyDecoder, MSONable, jsanitize
from monty.os import cd

from pydirac import SETTINGS
from pydirac.core.molecule import Molecule
from pydirac.core.periodic import Element
from pydirac.core.settings import Settings

__all__ = ["DiracInput", "Inp", "Mol"]


class DiracInput(dict, MSONable):
    """
    Class to contain a set of DIRAC input objects corresponding to a run.

    Parameters
    ----------
    inp : Inp
        Inp object.
    mol : Mol
        Mol object.
    optional_files : dict, optional
        Other input files supplied as a dictionary of {filename: object}. The object
        should follow standard conventions in implementing an `as_dict()` and
        `from_dict()` method. Defaults to None.

    Attributes
    ----------
    keys : List[str]
        List of keys in the DiracInput dictionary.

    """

    def __init__(self, inp, mol, optional_files=None, **kwargs):
        """
        Parameters
        ----------
        inp : Inp
            Inp object.
        mol : Mol
            Mol object.
        optional_files : dict, optional
            Other input files supplied as a dictionary of {filename: object}. The object
            should follow standard conventions in implementing an `as_dict()` and
            `from_dict()` method. Defaults to None.
        **kwargs
            Keyword arguments to be passed to the base class.

        """
        super().__init__(**kwargs)
        self.update({"inp": inp, "mol": mol})
        if optional_files is not None:
            self.update(optional_files)

    def __str__(self):
        """
        Returns a string representation of the DiracInput object.

        Returns
        -------
        str
            String representation of the DiracInput object.

        """
        output = []
        for k, v in self.items():
            output.append(k)
            output.append(str(v))
            output.append("")
        return "\n".join(output)

    def as_dict(self) -> dict:
        """
        Returns a MSONable dict representation of the DiracInput object.

        Returns
        -------
        dict
            MSONable dict representation of the DiracInput object.

        """
        d = {k: v.as_dict() for k, v in self.items()}
        d["@module"] = self.__class__.__module__
        d["@class"] = self.__class__.__name__
        return d

    @classmethod
    def from_dict(cls, d):
        """
        Returns a DiracInput object from a dictionary.

        Parameters
        ----------
        d : dict
            Dict representation.

        Returns
        -------
        DiracInput
            DiracInput object.

        """
        dec = MontyDecoder()
        sub_d = {"optional_files": {}}
        for k, v in d.items():
            # TODO: here, inp and mol is not filename but suffix
            if k in ["inp", "mol"]:
                sub_d[k.lower()] = dec.process_decoded(v)
            elif k not in ["@module", "@class"]:
                sub_d["optional_files"][k] = dec.process_decoded(v)
        return cls(**sub_d)

    def write_input(self, output_dir=".", make_dir_if_not_present=True):
        """
        Write DIRAC input to a directory.

        Parameters
        ----------
        output_dir : str, optional
            Directory to write to. Defaults to current directory (".").
        make_dir_if_not_present : bool, optional
            Create the directory if not present. Defaults to True.

        Returns
        -------
        None

        """
        if make_dir_if_not_present and not os.path.exists(output_dir):
            os.makedirs(output_dir)
        for k, v in self.items():
            if v is not None:
                with open(os.path.join(output_dir, k), "w") as f:
                    f.write(v.__str__())

    @staticmethod
    def from_directory(input_dir, optional_files=None):
        """
        Read in a set of DIRAC input from a directory. Note that only the standard
        inp, mol files are read unless optional_files is specified.

        Parameters
        ----------
        input_dir : str
            Directory to read DIRAC input from.
        optional_files : dict, optional
            Optional files to read in as well as a dictionary of {filename: Object type}.
            Object type must have a static method from_file. Defaults to None.

        Returns
        -------
        DiracInput
            DiracInput object.

        """
        sub_d = {}
        for fname, ftype in [("PYDIRAC.inp", Inp), ("PYDIRAC.mol", Mol)]:
            try:
                fullpath = os.path.join(input_dir, fname)
                sub_d[fname.lower()] = ftype.from_file(fullpath)
            except FileNotFoundError:
                sub_d[fname.lower()] = None

        sub_d["optional_files"] = {}
        if optional_files is not None:
            for fname, ftype in optional_files.items():
                sub_d["optional_files"][fname] = ftype.from_file(os.path.join(input_dir, fname))
        return DiracInput(**sub_d)

    def run_dirac(
        self,
        run_dir=".",
        dirac_cmd: list = None,
        output_file="dirac.out",
        err_file="dirac.err",
    ):
        """
        Write input files and run DIRAC.

        Parameters
        ----------
        run_dir : str, optional
            Where to write input files and do the run.
        dirac_cmd : list, optional
            Args to be supplied to run DIRAC.
        output_file : str, optional
            File to write output. Defaults to "dirac.out".
        err_file : str, optional
            File to write error. Defaults to "dirac.err".

        Returns
        -------
        None

        Raises
        ------
        RuntimeError
            If `dirac_cmd` is not supplied or `DIRAC_EXE` is not set in `.pydirac.yaml`.

        """
        self.write_input(output_dir=run_dir)
        dirac_cmd = dirac_cmd or SETTINGS.get("DIRAC_EXE")
        dirac_cmd = [os.path.expanduser(os.path.expandvars(t)) for t in dirac_cmd]
        if not dirac_cmd:
            raise RuntimeError(
                "You need to supply dirac_cmd or set the DIRAC_EXE in .pydirac.yaml to run DIRAC."
            )
        with cd(run_dir):
            with open(output_file, "w") as f_std, open(err_file, "w", buffering=1) as f_err:
                subprocess.check_call(dirac_cmd, stdout=f_std, stderr=f_err)


class Inp(dict, MSONable):
    """Class to represent DIRAC input file.

    Parameters
    ----------
    params : dict, optional
        Dictionary containing the input file data.

    Attributes
    ----------
    _allowed_duplicated_list : list
        A list of keywords that are allowed to be duplicated.
    _top : list
        A list of top-level keywords in the input file.

    Methods
    -------
    as_dict() -> dict
        Convert the instance to a dictionary.
    from_dict(d: dict) -> Inp
        Initialize the instance from a dictionary.
    add_sub_node(settinging_obj, dir_node, subdir_node, keyword_node, value_list) -> bool
        Add sub node to parent node.
    from_string(lines: list) -> Inp
        Initialize the instance from a list of strings.
    get_string() -> str
        Transform all contents of ``input`` branch of ``settings`` into string with blocks,
        subblocks, keys and values.
    __str__() -> str
        Return the string representation of the instance.

    Properties
    ----------
    use_wavefunc : bool
        True if wavefunction is to be used in the calculation, False otherwise.
    has_hamiltonian : bool
        True if Hamiltonian is present in the input file, False otherwise.
    hamiltonian : dict
        Dictionary representing the Hamiltonian data in the input file.
    calc_type : str
        The calculation type (e.g., 'energy', 'gradient', 'hessian', etc.) in the input file.
    calc_method : str
        The calculation method (e.g., 'HF', 'DFT', 'MP2', etc.) in the input file.
    electric_field : list
        List of electric field strengths and orientations in the input file.
    ci_calc_orbit : dict
        Dictionary representing the CI calculation orbit in the input file.
    wf_tag : str
        The wavefunction tag in the input file.

    Raises
    ------
    AssertionError
        If the directory node is None or if the value list is not a list.
        RuntimeError: If a duplicated key is encountered.

    Examples
    --------
    >>> inp = Inp({'dirac': {}})
    >>> inp
    {'dirac': {}}

    Notes
    -----
    The Inp class is used to represent DIRAC input files.

    """

    _allowed_duplicated_list = ["CIROOTS"]
    _top = ["dirac", "analyze", "hamiltonian", "integrals", "general"]

    def __init__(self, params=None):
        """Initialize the Inp class instance."""
        super().__init__()
        if params:
            pass
        self.update(params)
        self._parse_input()

    def as_dict(self) -> dict:
        """Convert the instance to a dictionary."""
        d = dict(self)
        d["@module"] = self.__class__.__module__
        d["@class"] = self.__class__.__name__
        return d

    @classmethod
    def from_dict(cls, d):
        """Initialize the instance from a dictionary."""
        return Inp({k: v for k, v in d.items() if k not in ("@module", "@class")})

    @staticmethod
    def add_sub_node(settinging_obj, dir_node, subdir_node, keyword_node, value_list):
        """Add sub node to parent node."""
        assert type(dir_node) is not None
        assert type(value_list) == list

        # whether add value to parent node
        # if yes: value_list has been added to parent node
        # if no: nothing has been done
        is_set = False

        if subdir_node is not None and keyword_node is not None:
            # -------------------------------------------------------
            # Case 1, for example:
            # -------------------------------------------------------
            # **ANALYZE
            # .MULPOP
            # *MULPOP
            # .VECPOP
            # 1..oo
            if len(value_list) > 0:
                settinging_obj[dir_node][subdir_node][keyword_node] = value_list
            else:
                settinging_obj[dir_node][subdir_node][keyword_node] = True
            is_set = True
        elif subdir_node is None and keyword_node is not None:
            # -------------------------------------------------------
            # Case 2, for example:
            # -------------------------------------------------------
            # **DIRAC
            # .TITLE
            # B, DOSSSS, KRCI
            if len(value_list) > 0:
                settinging_obj[dir_node][keyword_node] = value_list
            else:
                settinging_obj[dir_node][keyword_node] = True
            is_set = True
        elif subdir_node is not None and keyword_node is None:
            # if subdir_node is not None and keyword_node is None
            # -------------------------------------------------------
            # Case 3, for example:
            # -------------------------------------------------------
            # **INTEGRALS
            # *READINP
            # .UNCONTRACT
            #
            # when it's processing .UNCONTRACT, it's the case 3
            pass
        else:
            # if subdir_node is None and keyword_node is None
            # -------------------------------------------------------
            # Case 4, for example:
            # -------------------------------------------------------
            # **DIRAC
            #
            # when processing the part in which there is only directory
            pass
        return is_set

    @classmethod
    def from_string(cls, lines):
        """
        Initialize the instance from a list of strings.

        Parameters
        ----------
        lines : list
            List of strings representing the input file.

        Returns
        -------
        Inp
            Inp instance initialized from the input file.
        """
        setting = {}

        old_dir = None
        old_subdir = None
        old_dotkey = None
        curr_value_list = []

        for line in lines:
            if line.startswith("#"):
                # this is a comment line
                continue

            if line.startswith("**"):
                if old_dir is not None:
                    is_set = Inp.add_sub_node(
                        setting, old_dir, old_subdir, old_dotkey, curr_value_list
                    )
                    if is_set:
                        curr_value_list = []
                if len(curr_value_list) > 0:
                    Inp.add_sub_node(setting, old_dir, old_subdir, old_dotkey, curr_value_list)

                # this is a directory
                dir_name = line.lstrip("**").rstrip()
                old_dir = dir_name
                # setting[old_dir] = Settings()
                setting[old_dir] = {}

                ## clear current value for subdir and keyword
                old_subdir = None
                old_dotkey = None
                curr_value_list.clear()

            elif line.startswith("*"):
                # this is a sub-directory
                if old_dir is not None:
                    is_set = Inp.add_sub_node(
                        setting, old_dir, old_subdir, old_dotkey, curr_value_list
                    )
                    if is_set:
                        curr_value_list = []

                if line.startswith("*END"):
                    continue

                sub_dir_name = line.lstrip("*").rstrip()
                old_subdir = sub_dir_name

                assert old_dir is not None
                if sub_dir_name in setting[old_dir]:
                    ## if sub_dir_name in old_dir word list
                    setting[old_dir][sub_dir_name] = {}
                    setting[old_dir][sub_dir_name]["_en"] = True
                else:
                    ## this is a real second level directory
                    setting[old_dir][old_subdir] = {}

                ## clear current value for keyword
                old_dotkey = None
                curr_value_list = []

            elif line.startswith("."):
                # this is dot keyword
                curr_dotkey = line.lstrip(".").rstrip()
                if old_dir is not None:
                    ## first we probabily have two same dotkey, e.g., CIROOTS
                    is_set = Inp.add_sub_node(
                        setting, old_dir, old_subdir, old_dotkey, curr_value_list
                    )
                    if is_set:
                        curr_value_list = []

                    if (
                        old_subdir is not None
                        and curr_dotkey in setting[old_dir][old_subdir].keys()
                        and curr_dotkey.upper() in Inp._allowed_duplicated_list
                    ):
                        # first we need to clear curr_value_list and add this to the
                        # previous one
                        if is_set:
                            # curr_value_list = setting[old_dir][old_subdir][
                            #     curr_dotkey]
                            # curr_value_list.append('.' + curr_dotkey)

                            ## get id:
                            _tmp_k = []
                            for _k in setting[old_dir][old_subdir].keys():
                                if curr_dotkey == _k:
                                    continue
                                if curr_dotkey in _k:
                                    _tmp_k.append(int(_k.split("_id_")[-1]))
                            if len(_tmp_k):
                                new_id = max(_tmp_k) + 1
                            else:
                                new_id = 1
                            curr_dotkey = curr_dotkey + "_id_" + str(new_id)

                        old_dotkey = curr_dotkey
                        continue

                    # if old_subdir is None and curr_dotkey in setting[old_dir]:
                    #     print('{}, {}, {}'.format(old_dir, old_subdir,
                    #                               curr_dotkey))
                    #     curr_value_list.append('.' + curr_dotkey)
                    #     continue

                old_dotkey = curr_dotkey
                assert not (old_dir is None and old_subdir is None)

                ## clear value for value list
                curr_value_list = []
            else:
                # this is value of keyword
                curr_value = line.rstrip()
                if len(curr_value):
                    curr_value_list.append(curr_value)
        return Inp.from_dict(setting)

    @classmethod
    def from_file(cls, filename):
        """
        Parse DIRAC input file to restore python Settings object
        """
        with open(filename) as f:
            lines = f.readlines()
        return Inp.from_string(lines)

    def get_string(self):
        """Transform all contents of ``input`` branch of ``settings``
        into string with blocks, subblocks, keys and values.

        On the highest level alphabetic order of iteration is modified:
        keys occuring in class attribute ``_top`` are printed first.
        See :ref:`dirac-input` for details.
        """
        is_empty = lambda x: isinstance(x, dict) and len(x) == 0

        def parse_key(key, value):
            id_ptn = re.compile(r"^(.*)_id_\d+$")
            m = id_ptn.match(key)
            if m:
                key = m.group(1)
                if key.upper() not in self._allowed_duplicated_list:
                    raise RuntimeError("Duplicated key does not support!")
            if value is False:
                return ""
            ret = "." + key.upper() + "\n"
            if not (value is True or is_empty(value)):
                if isinstance(value, list):
                    for i in value:
                        ret += str(i) + "\n"
                else:
                    ret += str(value) + "\n"
            return ret

        def parse_block(block):
            enabler = "_en"
            ret = "**" + block.upper() + "\n"
            s = self[block]
            for k, v in s.items():
                if not isinstance(v, dict) or is_empty(v):
                    ret += parse_key(k, v)
            for k, v in s.items():
                if isinstance(v, dict) and enabler in v:
                    ret += parse_key(k, v[enabler])
            for k, v in s.items():
                if isinstance(v, dict) and len(v) > 0:
                    ret += "*" + k.upper() + "\n"
                    for kk, vv in v.items():
                        if kk != enabler:
                            ret += parse_key(kk, vv)
            return ret

        inp = ""
        for block in self._top:
            if block in self:
                inp += parse_block(block)
        for block in self:
            if block.lower() not in self._top:
                inp += parse_block(block)
        inp += "*END OF INPUT\n"
        return inp

    def __str__(self):
        """
        Return a string representation of the object.

        Returns
        -------
        str
            The string representation of the object.
        """
        return self.get_string()

    @property
    def use_wavefunc(self):
        """
        Return whether or not to use a wave function.

        Returns
        -------
        bool
            Whether or not to use a wave function.
        """
        return self._use_wavefunc

    @property
    def has_hamiltonian(self):
        """
        Return whether or not a Hamiltonian is present.

        Returns
        -------
        bool
            Whether or not a Hamiltonian is present.
        """
        return self._has_hamiltonian

    @property
    def hamiltonian(self):
        """
        Return the Hamiltonian.

        Returns
        -------
        str
            The Hamiltonian.
        """
        return self._hamiltonian

    @property
    def calc_type(self):
        """
        Return the calculation type.

        Returns
        -------
        str
            The calculation type.
        """
        return self._calc_type

    @property
    def calc_method(self):
        """
        Return the calculation method.

        Returns
        -------
        str
            The calculation method.
        """
        return self._calc_method

    @property
    def electric_field(self):
        """
        Return the strength of the electric field.

        Returns
        -------
        float
            The strength of the electric field.
        """
        return self._electric_field

    @property
    def ci_calc_orbit(self):
        """
        Return the CI calculation orbit.

        Returns
        -------
        dict
            The CI calculation orbit.
        """
        if hasattr(self, "_ci_calc_orbit"):
            return self._ci_calc_orbit
        return {"occ": 0, "vir": 0}

    def _find_real_key(self, k, keys):
        """
        Find the real key from a list of keys.

        Parameters
        ----------
        k : str
            The key to find.
        keys : list
            The list of keys to search.

        Returns
        -------
        str or None
            The real key if found, otherwise None.
        """
        for _k in keys:
            if k in _k:
                return _k
        else:
            return None

    @property
    def wf_tag(self):
        """
        Return the wave function tag.

        Returns:
        str
             the wave function tag.
        """
        return self._wf_tag

    def _parse_input(self):
        """Parse input from output file

        Returns
        -------
            input settings
        """

        inp_settings = Settings(self)

        self._wf_tag = None
        for dk in inp_settings.dirac:
            if "WAVE F" in dk:
                self._use_wavefunc = True
                self._wf_tag = self._find_real_key("WAVE F", self)
                break
        else:
            self._use_wavefunc = False
            self._has_hamiltonian = False
            self._hamiltonian = None
            self._calc_type = None
            self._calc_method = None
            self._electric_field = None
            return

        # check Hamiltonian
        if "HAMILTONIAN" in inp_settings:
            self._has_hamiltonian = True
            if "DOSSSS" in inp_settings.hamiltonian:
                # this is 4-component Hamiltonian
                self._hamiltonian = "4C"
            elif "X2C" in inp_settings.hamiltonian:
                # this is 2-component Hamiltonian
                self._hamiltonian = "2C"
            elif "X2Cmmf" in inp_settings.hamiltonian:
                # need to do
                self._hamiltonian = "2C"
            elif "NONREL" in inp_settings.hamiltonian:
                # self._hamiltonian = "NR-2C"
                self._hamiltonian = "2C-NR"
            else:
                warnings.warn("We do not know how many components, " "please check inp file")
                self._hamiltonian = "?C"

            if not "NONREL" in inp_settings.hamiltonian:
                if "NOSPIN" in inp_settings.hamiltonian:
                    self._hamiltonian += "-SR"
                else:
                    self._hamiltonian += "-DC"

            if "Gaunt" in inp_settings.hamiltonian:
                self._hamiltonian = self._hamiltonian + "-Gaunt"
        else:
            self._has_hamiltonian = False

        # check if CC or CI calculation
        if (
            self._has_hamiltonian
            and "RELCC" in inp_settings
            and "RELCCSD" in inp_settings[self._wf_tag]
        ):
            self._calc_method = "CC"
        elif (
            self._has_hamiltonian
            and "KRCICALC" in inp_settings[self._wf_tag]
            and "KR CI" in inp_settings[self._wf_tag]
        ):
            self._calc_method = "CI"
            # here, we can obtain calc_orbit info from input directly
            gas_tag = None
            for k in inp_settings[self._wf_tag]["KRCICALC"].keys():
                if "GAS" in k:
                    gas_tag = k
                    break
            gas_res = inp_settings[self._wf_tag]["KRCICALC"][gas_tag]
            # the first line is the number of GAS shell
            gas_nb = int(gas_res[0])
            assert gas_nb == len(gas_res[1:])

            nb_e = 0
            t_orb = []
            gas_ptn = re.compile(r"\s*\d+\s+(?P<nb_e>\d+)\s*/\s*(?P<nb_orb>\d+)")
            for l in gas_res[1:]:
                m = gas_ptn.match(l)
                if m:
                    t_orb.append(int(m.group("nb_orb")))
            m = gas_ptn.match(gas_res[-1])
            if m:
                nb_e = int(m.group("nb_e"))
                tot_orb = sum(t_orb)
                occ_orb = nb_e
                vir_orb = tot_orb * 2 - occ_orb
                self._ci_calc_orbit = {"occ": occ_orb, "vir": vir_orb}
            else:
                self._ci_calc_orbit = {"occ": 0, "vir": 0}

        elif self._has_hamiltonian and "SCF" in inp_settings[self._wf_tag]:
            if inp_settings[self._wf_tag]["SCF"]["_en"]:
                self._calc_method = "SCF"
            else:
                self._calc_method = None
        # TODO: visual module calculation
        else:
            self._calc_method = None

        # check dipole or quadrupole
        if self._has_hamiltonian and "OPERATOR" in inp_settings.hamiltonian:
            if " ZZTHETA" in inp_settings.hamiltonian.operator:
                self._calc_type = "Q"
            elif " ZDIPLEN" in inp_settings.hamiltonian.operator:
                self._calc_type = "D"
            else:
                warnings.warn("this is not finite-field calculations")
                self._calc_type = "non_finite_field"
        else:
            if not self._has_hamiltonian:
                warnings.warn("Hamiltonian is invalid")
            elif not "OPERATOR" in inp_settings.hamiltonian:
                # warnings.warn('".OPERATOR" is not in Hamiltonian, this is not finit-field calculation!')
                pass
            self._calc_type = "non_finite_field"

        # check electric field
        if self._calc_type in "QD":
            # this is finit-field
            pos = inp_settings.hamiltonian.operator.index(" COMFACTOR")
            self._electric_field = inp_settings.hamiltonian.operator[pos + 1]
        else:
            # warnings.warn('Did not find electric field')
            self._electric_field = "null"

    def find_case(self, key):
        """Check if this instance contains a key consisting of the same
        letters as *key*, but possibly with different case. If found, return
        such a key. If not, return *key*.

        Args:
            key:

        Returns:

        """
        if not isinstance(key, str):
            return key
        lowkey = key.lower()
        for k in self:
            if k.lower() == lowkey:
                return k
        return key

    def __contains__(self, name):
        """Like regular ``__contains`__``, but ignore the case."""
        return dict.__contains__(self, self.find_case(name))

    def __getitem__(self, name):
        """Like regular ``__getitem__``, but ignore the case."""
        return dict.__getitem__(self, self.find_case(name))

    def __setitem__(self, name, value):
        """Like regular ``__setitem__``, but ignore the case and if the
        value is a dict, convert it to |Settings|.
        """
        dict.__setitem__(self, self.find_case(name), value)

    def __delitem__(self, name):
        """Like regular ``__detitem__``, but ignore the case."""
        return dict.__delitem__(self, self.find_case(name))

    def write_file(self, filename):
        with open(filename, "w") as f:
            f.write(self.__str__())


class Mol(MSONable):
    """Class to represent DIRAC Mol file"""

    _basis_info_line = 2

    def __init__(
        self,
        molecule: Molecule,
        comment: str = None,
        basis_type: str = None,
        basis_lib: str = "BASIS",
    ):
        self.molecule = molecule
        self.comment = comment
        self.basis_type = basis_type
        self.basis_lib = basis_lib

    def as_dict(self) -> dict:
        """Return object as dict."""
        d = {
            "molecule": self.molecule.as_dict(),
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
        }

        return d

    @classmethod
    def from_file(cls, filename):
        """Initialize from file

        Parameters
        ----------
        filename: str
            Filename
        """
        # TODO, more info from mol file including basis sets.
        with open(filename) as f:
            lines = f.readlines()
        return Mol.from_string(lines)

    @classmethod
    def from_string(cls, context):
        """
        Construct a `Mol` object by parsing the contents of a MOL file.

        Parameters
        ----------
        context : list of str
            The contents of the MOL file as a list of strings.

        Returns
        -------
        Mol
            A `Mol` object representing the molecule.

        """

        start_line = re.compile(r"^Contents of the molecule file\s*")
        dash_line = re.compile(r"^-+")
        empty_line = re.compile(r"^$")
        # inp_start_line = re.compile(r"^\*\*DIRAC\s*")
        nuclei_nb_line = re.compile(
            r"^\s*(?P<nuclei>[-+]?(\d+(\.\d*)?|\d*\.\d+))\s+(?P<nb_atoms>\d+)\s*$"
        )
        basis_line = re.compile(r"^\b(?:LARGE|EXPLICIT)\b\s+BASIS\s+(\S+)\s*")
        endline = re.compile(r"^FINISH")

        nuclei_id = 0
        nb_atoms = 0
        basis_type = None

        for i, line in enumerate(context):
            if endline.match(line):
                break
            elif dash_line.match(line) or empty_line.match(line):
                continue
            elif nuclei_nb_line.match(line):
                m = nuclei_nb_line.match(line)
                nuclei_id = int(round(float(m.group("nuclei"))))
                nb_atoms = int(m.group("nb_atoms"))
                # print(self.nuclei_id, self.nb_atoms)
            elif basis_line.match(line):
                m = basis_line.match(line)
                if m:
                    basis_type = m.group(1)
                    break
            else:
                # we do not interest with these lines
                continue

        if basis_type is None:
            basis_type = context[Mol._basis_info_line].strip()
        basis_type = basis_type
        # TODO: we are doing atomic calculation
        assert nb_atoms == 1
        molecule = Molecule(atoms=[nuclei_id], coordinates=[[0.0, 0.0, 0.0]])
        return cls(basis_type=basis_type, molecule=molecule)

    @staticmethod
    def get_string(
        atom_info: Element,
        basis_type: str,
        basis_lib: str = "BASIS",
    ) -> str:
        """
        Return a string representation of a molecule in MOL format with the specified
        atom information and basis.

        Parameters
        ----------
        atom_info : Element
            The atomic information for the molecule.
        basis_type : str
            The type of basis.
        basis_lib : str, optional
            The type of basis library. Should be either "BASIS" or "EXPLICIT".

        Returns
        -------
        str
            A string representation of the molecule in MOL format.

        Raises
        ------
        TypeError
            If the `basis_lib` parameter is not "BASIS" or "EXPLICIT".

        """
        atom_index = atom_info.Z
        atom_type = atom_info.symbol

        if basis_lib not in ["EXPLICIT", "BASIS"]:
            raise TypeError(
                'Basis type should be "BASIS" or "EXPLICIT" ' "for builtin basis or custom basis."
            )

        if basis_lib == "EXPLICIT":
            with open(f"basis/{atom_type}.dat") as f:
                basis_info = f.read()
            template = Mol.get_mol_by_custom_basis(atom_type, atom_index, basis_lib, basis_info)
        elif basis_lib == "BASIS":
            template = Mol.get_mol_by_default_basis(atom_type, atom_index, basis_type)
        else:
            raise TypeError(
                'Basis type should be "BASIS" or "EXPLICIT" ' "for builtin basis or custom basis."
            )
        return template

    @staticmethod
    def get_mol_by_custom_basis(atom_type, atom_index, basis_choice, basis_info):
        """
        Return a string representation of a molecule in MOL format with a custom basis.

        Parameters
        ----------
        atom_type : str
            The type of atom.
        atom_index : int
            The index of the atom.
        basis_choice : str
            The type of basis.
        basis_info : str
            The custom basis information.

        Returns
        -------
        str
            A string representation of the molecule in MOL format with the custom basis.

        """
        template = """INTGRL
{atom} atom
{basis}
C   1    2  X  Y
    {elec}.    1
{atom}    0.0000000000        0.0000000000        0.0000000000
{basis_info}
""".format(
            **{
                "atom": atom_type,
                "elec": atom_index,
                "basis": basis_choice,
                "basis_info": basis_info,
            }
        )
        return template

    @staticmethod
    def get_mol_by_default_basis(atom_type, atom_index, basis_type):
        """
        Return a string representation of a molecule in MOL format, with the
        specified atom type, atom index, and basis type.

        Parameters
        ----------
        atom_type : str
            The type of atom.
        atom_index : int
            The index of the atom.
        basis_type : str
            The type of basis.

        Returns
        -------
        str
            A string representation of the molecule in MOL format.

        """
        template = """INTGRL
{atom} atom
{basis}
C   1    2  X  Y
    {elec}.    1
{atom}    0.0000000000        0.0000000000        0.0000000000
LARGE BASIS {basis}
FINISH

""".format(
            **{"atom": atom_type, "elec": atom_index, "basis": basis_type}
        )
        return template

    def __str__(self):
        """
        Return a string representation of the molecule.

        Returns
        -------
        str
            A string representation of the molecule.

        """

        basis_type = self.basis_type or "null"
        if self.basis_lib == "EXPLICIT":
            warnings.warn(f"No implement {self.basis_lib} yet")
            pass
        return Mol.get_string(self.molecule.atomic_info, basis_type)

    def write_file(self, filename=None):
        """
        Write the molecule to a file in MOL format.

        Parameters
        ----------
        filename : str, optional
            The name of the file to write to. If not provided, the file will be
            named after the atom symbol and basis type.

        Returns
        -------
        None

        Raises
        ------
        NotImplementedError
            If the molecule contains more than one atom.

        """
        if self.molecule.is_atom:
            fname = filename or self.molecule.atomic_info.symbol + "_" + self.basis_type + ".mol"
            with open(fname, "w") as f:
                basis_type = self.basis_type or "null"
                f.write(Mol.get_string(self.molecule.atomic_info, basis_type))
        else:
            raise NotImplementedError(
                "For many-atoms molecule, " "it has not been implemented yet! "
            )

    def update(self, new_dict):
        """
        Update the attributes of the molecule with the values in `new_dict`.

        Parameters
        ----------
        new_dict : dict
            A dictionary of attribute-value pairs to update.

        Returns
        -------
        None

        """
        for k, v in new_dict.items():
            setattr(self, k, v)



import glob
import os
import warnings

import numpy as np
from monty.os import cd

from pydirac.analysis.utility import get_keyword, get_orbital_info
from pydirac.io.outputs import Output

__all__ = [
    "PolarizabilityCalculator",
    "get_polarizability",
    "get_polarizability_from_output_list",
    "do_one_basis",
    "is_valid",
]


class PolarizabilityCalculator:
    """
    A class to calculate dipole polarizability and quadrupole polarizability using finite difference
    method.

    Parameters
    ----------
    calc_type : str, optional
        The type of calculation to perform. Must be one of "dipole" or "quadrupole".
        Default is "dipole".
    nb_coeffs : int, optional
        The number of coefficients to use for the calculation. Must be 2 or 3.
        Default is 2.

    Attributes
    ----------
    calc_type : str
        The type of calculation to perform. Must be one of "dipole" or "quadrupole".
    nb_coeffs : int
        The number of coefficients to use for the calculation.
    """

    def __init__(self, calc_type="dipole", nb_coeffs=2):
        """
        Initialize the PolarizabilityCalculator instance.

        Parameters
        ----------
        calc_type : str, optional
            The type of calculation to perform. Must be one of "dipole" or "quadrupole".
            Default is "dipole".
        nb_coeffs : int, optional
            The number of coefficients to use for the calculation. Must be 2 or 3.
            Default is 2.
        """
        _ct = calc_type.lower()
        if _ct in ["d", "dipole"]:
            _ct = "dipole"
        elif _ct in ["q", "quad", "quadrupole"]:
            _ct = "quadrupole"
        else:
            raise ValueError("Wrong calculation type!")
        self.calc_type = _ct

        if nb_coeffs not in (2, 3):
            raise ValueError("nb_coeffs must be 2 or 3!")
        self.nb_coeffs = nb_coeffs

    def get_coeff(self, f):
        """
        Return the coefficients for a given electric field strength.

        Parameters
        ----------
        f : float
            The electric field strength.

        Returns
        -------
        array_like of float
            The coefficients.
        """
        if self.nb_coeffs == 2:
            if self.calc_type == "dipole":
                return np.array([-np.power(f, 2) / 2.0, -np.power(f, 4) / 24.0])
            elif self.calc_type == "quadrupole":
                return np.array([np.power(f, 1) / 2.0, -np.power(f, 2) / 8.0])
            else:
                raise TypeError(
                    'calc_type {} is not valid, please use "dipole" '
                    'or "quadrupole"'.format(self.calc_type)
                )

        elif self.nb_coeffs == 3:
            if self.calc_type == "dipole":
                return np.array(
                    [
                        -np.power(f, 2) / 2.0,
                        -np.power(f, 4) / 24.0,
                        -np.power(f, 3) / 144.0,
                    ]
                )
            elif self.calc_type == "quadrupole":
                return np.array(
                    [
                        -np.power(f, 1) / 2.0,
                        -np.power(f, 2) / 8.0,
                        -np.power(f, 3) / 24.0,
                    ]
                )
            else:
                raise TypeError(
                    'calc_type {} is not valid, please use "dipole" '
                    'or "quadrupole"'.format(self.calc_type)
                )
        else:
            raise ValueError(
                f"calc_type {self.calc_type} is not valid, please use 'dipole' or 'quadrupole'."
            )

    def get_A(self, field):
        """
        Create the matrix A for the given electric field strengths.

        Parameters
        ----------
        field : array_like of float
            The electric field strengths.

        Returns
        -------
        array_like of float, shape (n_samples, nb_coeffs)
            The matrix A for the given electric field strengths.
        """
        A = np.zeros((len(field), self.nb_coeffs))
        for i, v in enumerate(field):
            A[i, :] = self.get_coeff(v)
        return A

    def get_svd_from_array(self, energy, field, rcond=None, threshold=0.005):
        """
        Calculate the SVD-based solution for the least-squares problem.

        Parameters
        ----------
        energy : array_like of float
            The energy values.
        field : array_like of float
            The electric field strengths.

        Returns
        -------
        array_like of float
            The coefficients
        """
        field = np.asarray(field)
        energy = np.asarray(energy)
        mask = field <= threshold + 1e-8
        field = field[mask]
        energy = energy[mask]

        e_ref = 0.0
        for e, f in zip(energy, field):
            if np.isclose(f, 0.0):
                e_ref = e

        # define a matrix
        A = self.get_A(np.asarray(field))
        b = np.asarray(energy) - e_ref

        # x_svd = np.linalg.lstsq(A, b, rcond=1e-8)[0]
        # x_svd = np.linalg.lstsq(A, b, rcond=None)[0]
        # print(x_svd)

        # Solve the least squares problem using SVD
        x_svd, residuals, rank, s = np.linalg.lstsq(A, b, rcond=rcond)

        # Manually calculate residuals if not provided by np.linalg.lstsq
        if residuals.size == 0:
            residuals = b - np.dot(A, x_svd)
            residual_sum_of_squares = np.sum(residuals**2)
        else:
            residual_sum_of_squares = residuals[0]
            # residuals2 = b - np.dot(A, x_svd)
            # residual_sum_of_squares2 = np.sum(residuals2**2)
            # assert np.allclose(residual_sum_of_squares, residual_sum_of_squares2)
            # print("assert True")

        # Calculate the covariance matrix of the coefficients
        # Assuming homoscedasticity (constant variance of residuals)
        MSE = residual_sum_of_squares / (len(energy) - len(x_svd))  # Mean Squared Error
        cov_matrix = np.linalg.inv(np.dot(A.T, A)) * MSE

        # Standard errors are the square roots of the diagonal elements of the covariance matrix
        standard_errors = np.sqrt(np.diag(cov_matrix))
        return x_svd, standard_errors

    # def get_res_svd(self, filename):
    #     """
    #     Calculate the SVD-based solution for the least-squares problem using the data from a file.

    #     Parameters
    #     ----------
    #     filename : str
    #         The name of the file containing the electric field strengths and energy values.

    #     Returns
    #     -------
    #     array_like of float
    #         The coefficients.
    #     """
    #     with open(filename) as f:
    #         context = f.readlines()

    #     if len(context) < 3:
    #         raise ValueError(f"Data points in {filename} is less than 3")

    #     filed_energy = [tuple(c.split()) for c in context]
    #     energy, field = [], []
    #     for f, e in filed_energy:
    #         f = float(f)
    #         energy.append(float(e))
    #         field.append(f)

    #     x_svd = self.get_svd_from_array(energy, field)
    #     return x_svd


def get_polarizability(
    dirname: str = "./", calc_dir_patters=None, deepth: int = 0, verbos=False, threshold=0.005
) -> dict:
    """Get polarizability from a directory

    A calculation directory may contain some information as followings:
        (1) no more calculation directory (calc_dir_list) but several output files (curr_dir_output_list)
        (2) several output files (curr_dir_output_list) with other calculation directories (calc_dir_list)
        (3) several output files (curr_dir_output_list) with subdirectories (sub_dir_list) --> call self again when deepth > 0
        (4) only calculations directories (calc_dir_list)
        (5) calculations directories (calc_dir_list) and (sub_dir_list)
        (6) only subdirectories (sub_dir_list)
        (7) all (calc_dir_list), (curr_dir_output_list) and (sub_dir_list)

    Args:
        dirname: dirname
        suffix:

    Returns:
        None
    """
    # deepth = 0: only calc current output files
    if calc_dir_patters is None:
        calc_dir_patters = ["*dyall*"]

    if isinstance(calc_dir_patters, (str, list)):
        if isinstance(calc_dir_patters, str):
            calc_dir_patters = [calc_dir_patters]
    else:
        raise TypeError("calc_dir_patters can only be <str> or <list>")

    deepth = 0 if deepth < 0 else deepth
    do_sub_dir = deepth > 0

    # for debug
    if verbos:
        print(f"cd {dirname}")

    all_res = {}
    curr_dir_output_lis = []
    calc_dir_output_lis = []
    with cd(dirname):
        # Step 1. deal with all current output files
        # ------------------------------
        for f in glob.glob("*.out"):
            obj = Output(filename=f)
            if obj.is_ok:
                curr_dir_output_lis.append(obj)

        # Step 2. deal with all calc_dir
        # ------------------------------
        clc_dirs = []
        for ptn in calc_dir_patters:
            clc_dirs.extend([d for d in glob.glob("*" + ptn + "*") if os.path.isdir(d)])
        # if has calc_dir then I would try
        for clc_d in clc_dirs:
            with cd(clc_d):
                outs = glob.glob("*.out")
                if len(outs) > 1:
                    warnings.warn(
                        "there are more than two output file in this directory {}, and we take "
                        "the first one {}".format(clc_d, outs[0])
                    )
                elif len(outs) == 1:
                    obj = Output(filename=outs[0])
                    if obj.is_ok:
                        calc_dir_output_lis.append(obj)
                else:
                    warnings.warn(f"there is no output file in this directory {clc_d}")

        all_res["curr_dir"] = {}
        if is_valid(curr_dir_output_lis):
            e_dict = get_polarizability_from_output_list(
                dirname, curr_dir_output_lis, tag="curr_dir", verbos=verbos, threshold=threshold
            )
            all_res["curr_dir"] = e_dict

        all_res["calc_dir"] = {}
        if is_valid(calc_dir_output_lis):
            e_dict = get_polarizability_from_output_list(
                dirname, calc_dir_output_lis, tag="clc_dir", verbos=verbos, threshold=threshold
            )
            all_res["calc_dir"] = e_dict

        # Step 3. deal with all sub_dir
        # ------------------------------
        all_res["sub_dir"] = {}
        if do_sub_dir:
            sub_dirs = [d for d in glob.glob("*") if os.path.isdir(d) and d not in clc_dirs]

            for sd in sub_dirs:
                res = get_polarizability(
                    os.path.join(dirname, sd),
                    calc_dir_patters,
                    deepth - 1,
                    verbos=verbos,
                    threshold=threshold,
                )
                if sd not in all_res["sub_dir"] and len(res) > 0:
                    all_res["sub_dir"][os.path.basename(sd)] = res
    return all_res


def do_one_basis(output_lis, threshold=0.005):
    if not len(output_lis):
        warnings.warn("there is no valid output file here")
        return

    # check the calc_type from the first output file
    ct = output_lis[0].inp.calc_type
    calc_orbit = output_lis[0].calc_orbit
    pc = PolarizabilityCalculator(calc_type=ct)

    # maybe we just use a dict to restore all information
    energies = {}
    # cause for different type calculations, the electric fields are the same
    fields = []

    for o in output_lis:
        if o.inp.calc_type != ct:
            warnings.warn(
                "we found a error output whose calc_type {} is "
                "different with the first one {}".format(o.inp.calc_type, ct)
            )
            continue

        # CC or CI
        fields.append(float(o.inp.electric_field))
        if o.inp.calc_method == "CC":
            for k, v in o.energy_settings.items():
                if k not in energies.keys():
                    energies[k] = []
                energies[k].append(v)

        elif o.inp.calc_method == "CI":
            # check all roots converged
            for i in o.energy_settings["ci_converged"]:
                if not np.all(np.asarray(i)):
                    raise RuntimeError("Not all roots are converged!")

            for k, v in o.energy_settings["ci_e"].items():
                if k not in energies.keys():
                    energies[k] = []
                energies[k].append(v)

            if "scf_e" not in energies:
                energies["scf_e"] = []
            energies["scf_e"].append(o.energy_settings["scf_e"])
        else:
            # raise NotImplementedError
            continue

    # remove items where the length of energy is not equal to the one of field
    del_k_lis = []
    for k, v in energies.items():
        if len(v) != len(fields):
            warnings.warn(
                'The length of energy set "{}" {} does '
                "not equal the length of field {}".format(k, len(v), len(fields))
            )
            del_k_lis.append(k)
    for k in del_k_lis:
        energies.pop(k)

    # {'scf': alpha_scf, 'mp2':alpha_scf, 'ccsd':alpha_ccsd}
    res = {}
    res_error = {}
    for k, energy in energies.items():
        polar_key = k.strip("_e")
        _polar, _error = pc.get_svd_from_array(energy, fields, threshold)
        res[polar_key] = _polar
        res_error[polar_key] = _error
    res_all = {
        "energy": {"energies": energies, "fields": fields},
        "polar": res,
        "polar_error": res_error,
    }
    return res_all


def get_polarizability_from_output_list(
    dirname, output_lis, tag=None, verbos=True, threshold=0.005
):
    all_basis_res = {}
    for o in output_lis:
        task_type, orbit = o.task_type, o.calc_orbit
        obt_info = get_orbital_info(orbit["occ"], orbit["vir"]) if len(orbit) else "null"
        k = get_keyword(o.mol.molecule.atomic_info.symbol, task_type, obt_info)
        if k not in all_basis_res:
            all_basis_res[k] = []
        all_basis_res[k].append(o)

    e_res = {}
    for k, v in all_basis_res.items():
        one_res = do_one_basis(v, threshold=threshold)
        if one_res:
            e_res[k] = one_res
    return e_res


def is_valid(output_lis, verbos=False):
    """Check whether a output_lis is valid

    Args:
        output_lis: a list of Output obj

    Returns:
        True or False
    """

    if len(output_lis) < 3:
        if verbos:
            warnings.warn("the nubmer of output objects is less than 3")
        return False

    # check all there file if they are all 'CC' or 'CI' calculations
    task_record = {}
    for o in output_lis:
        if o.inp.calc_method in ["CC", "CI"]:
            if not o.task_type in task_record:
                task_record[o.task_type] = 1
            else:
                task_record[o.task_type] += 1
    for v in task_record.values():
        if v >= 3:
            return True
    else:
        if verbos:
            print(task_record)
            warnings.warn("the maximum of output objects with the same type is less than 3")
        return False



# NOTE: This file is from the ACM Palm package. Please reference it if you use this
#       class in your project.

import contextlib
import textwrap
from functools import wraps

__all__ = ["Settings"]


class Settings(dict):
    """Automatic multi-level dictionary. Subclass of built-in class :class:`dict`.

    The shortcut dot notation (``s.basis`` instead of ``s['basis']``) can be used for keys that:

    *   are strings
    *   don't contain whitespaces
    *   begin with a letter or an underscore
    *   don't both begin and end with two or more underscores.

    Iteration follows lexicographical order (via :func:`sorted` function)

    Methods for displaying content (:meth:`~object.__str__` and
    :meth:`~object.__repr__`) are overridden to recursively show nested
    instances in easy-readable format.

    Regular dictionaries (also multi-level ones) used as values (or passed to
    the constructor) are automatically transformed to |Settings| instances::

        >>> s = Settings({'a': {1: 'a1', 2: 'a2'}, 'b': {1: 'b1', 2: 'b2'}})
        >>> s.a[3] = {'x': {12: 'q', 34: 'w'}, 'y': 7}
        >>> print(s)
        a:
          1:    a1
          2:    a2
          3:
            x:
              12:   q
              34:   w
            y:  7
        b:
          1:    b1
          2:    b2

    """

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        for k, v in self.items():
            if isinstance(v, dict) and not isinstance(v, Settings):
                self[k] = Settings(v)
            if isinstance(v, list):
                self[k] = [
                    Settings(i) if (isinstance(i, dict) and not isinstance(i, Settings)) else i
                    for i in v
                ]

    def copy(self):
        """Return a new instance that is a copy of this one. Nested |Settings|
        instances are copied recursively, not linked.

        In practice this method works as a shallow copy: all "proper values"
        (leaf nodes) in the returned copy point to the same objects as the
        original instance (unless they are immutable, like ``int`` or ``tuple``).
        However, nested |Settings| instances (internal nodes) are copied in a
        deep-copy fashion. In other words, copying a |Settings| instance creates
        a brand new "tree skeleton" and populates its leaf nodes with values
        taken directly from the original instance.

        This behavior is illustrated by the following example::

            >>> s = Settings()
            >>> s.a = 'string'
            >>> s.b = ['l','i','s','t']
            >>> s.x.y = 12
            >>> s.x.z = {'s','e','t'}
            >>> c = s.copy()
            >>> s.a += 'word'
            >>> s.b += [3]
            >>> s.x.u = 'new'
            >>> s.x.y += 10
            >>> s.x.z.add(1)
            >>> print(c)
            a:  string
            b:  ['l', 'i', 's', 't', 3]
            x:
              y:    12
              z:    set([1, 's', 'e', 't'])
            >>> print(s)
            a:  stringword
            b:  ['l', 'i', 's', 't', 3]
            x:
              u:    new
              y:    22
              z:    set([1, 's', 'e', 't'])

        This method is also used when :func:`python3:copy.copy` is called.
        """
        ret = Settings()
        for name in self:
            if isinstance(self[name], Settings):
                ret[name] = self[name].copy()
            else:
                ret[name] = self[name]
        return ret

    def soft_update(self, other):
        """Update this instance with data from *other*, but do not overwrite
        existing keys. Nested |Settings| instances are soft-updated recursively.

        In the following example ``s`` and ``o`` are previously prepared
        |Settings| instances::

            >>> print(s)
            a:  AA
            b:  BB
            x:
              y1:   XY1
              y2:   XY2
            >>> print(o)
            a:  O_AA
            c:  O_CC
            x:
              y1:   O_XY1
              y3:   O_XY3
            >>> s.soft_update(o)
            >>> print(s)
            a:  AA        #original value s.a not overwritten by o.a
            b:  BB
            c:  O_CC
            x:
              y1:   XY1   #original value s.x.y1 not overwritten by o.x.y1
              y2:   XY2
              y3:   O_XY3

        *Other* can also be a regular dictionary. Of course in that case only
        top level keys are updated.

        Shortcut ``A += B`` can be used instead of ``A.soft_update(B)``.
        """
        for name in other:
            if isinstance(other[name], Settings):
                if name not in self:
                    self[name] = other[name].copy()
                elif isinstance(self[name], Settings):
                    self[name].soft_update(other[name])
            elif name not in self:
                self[name] = other[name]
        return self

    def update(self, other):
        """Update this instance with data from *other*, overwriting existing keys.
        Nested |Settings| instances are updated recursively.

        In the following example ``s`` and ``o`` are previously prepared |Settings| instances::

            >>> print(s)
            a:  AA
            b:  BB
            x:
              y1:   XY1
              y2:   XY2
            >>> print(o)
            a:  O_AA
            c:  O_CC
            x:
              y1:   O_XY1
              y3:   O_XY3
            >>> s.update(o)
            >>> print(s)
            a:  O_AA        #original value s.a overwritten by o.a
            b:  BB
            c:  O_CC
            x:
              y1:   O_XY1   #original value s.x.y1 overwritten by o.x.y1
              y2:   XY2
              y3:   O_XY3

        *Other* can also be a regular dictionary.
        Of course in that case only top level keys are updated.
        """
        for name in other:
            if isinstance(other[name], Settings):
                if name not in self or not isinstance(self[name], Settings):
                    self[name] = other[name].copy()
                else:
                    self[name].update(other[name])
            else:
                self[name] = other[name]

    def merge(self, other):
        """Return new instance of |Settings| that is a copy of this
        instance soft-updated with *other*.

        Shortcut ``A + B`` can be used instead of ``A.merge(B)``.
        """
        ret = self.copy()
        ret.soft_update(other)
        return ret

    def find_case(self, key):
        """Check if this instance contains a key consisting of the same letters
        as *key*, but possibly with different case. If found, return such a key.
        If not, return *key*.
        """
        if not isinstance(key, str):
            return key
        lowkey = key.lower()
        for k in self:
            if k.lower() == lowkey:
                return k
        return key

    def as_dict(self):
        """Return a copy of this instance with all |Settings| replaced by
        regular Python dictionaries.
        """
        d = {}
        for k, v in self.items():
            if isinstance(v, Settings):
                d[k] = v.as_dict()
            elif isinstance(v, list):
                d[k] = [i.as_dict() if isinstance(i, Settings) else i for i in v]
            else:
                d[k] = v

        return d

    @classmethod
    def suppress_missing(cls):
        """A context manager for temporary disabling the :meth:`.Settings.__missing__`
        magic method: all calls now raising a :exc:`KeyError`.

        As a results, attempting to access keys absent from an arbitrary |Settings|
        instance will raise a :exc:`KeyError`, thus reverting to the default
        dictionary behaviour.

        .. note::
            The :meth:`.Settings.__missing__` method is (temporary) suppressed
            at the class level to ensure consistent invocation by the Python interpreter.
            See also `special method lookup`_.

        Example:

        .. code:: python

            >>> s = Settings()

            >>> with s.suppress_missing():
            ...     s.a.b.c = True
            KeyError: 'a'

            >>> s.a.b.c = True
            >>> print(s.a.b.c)
            True

        .. _`special method lookup`: https://docs.python.org/3/reference/datamodel.html#special-method-lookup

        """
        return SuppressMissing(cls)

    def get_nested(self, key_tuple, suppress_missing=False):
        """Retrieve a nested value by, recursively, iterating through this
        instance using the keys in *key_tuple*.

        The :meth:`.Settings.__getitem__` method is called recursively on this
        instance until all keys in key_tuple are exhausted.

        Setting *suppress_missing* to ``True`` will internally open the :meth:`.
        Settings.suppress_missing` context manager, thus raising a :exc:`KeyError`
        if a key in *key_tuple* is absent from this instance.

        .. code:: python

            >>> s = Settings()
            >>> s.a.b.c = True
            >>> value = s.get_nested(('a', 'b', 'c'))
            >>> print(value)
            True
        """
        s = self
        with contextlib.suppress() if not suppress_missing else s.suppress_missing():
            for k in key_tuple:
                s = s[k]
        return s

    def set_nested(self, key_tuple, value, suppress_missing=False):
        """Set a nested value by, recursively, iterating through this instance
        using the keys in *key_tuple*.

        The :meth:`.Settings.__getitem__` method is called recursively on this
        instance, followed by :meth:`.Settings.__setitem__`, until all keys in
        key_tuple are exhausted.


        Setting *suppress_missing* to ``True`` will internally open the :meth:`.
        Settings.suppress_missing` context manager, thus raising a :exc:`KeyError`
        if a key in *key_tuple* is absent from this instance.

        .. code:: python

            >>> s = Settings()
            >>> s.set_nested(('a', 'b', 'c'), True)
            >>> print(s)
            a:
              b:
                c: 	True
        """
        s = self
        with contextlib.suppress() if not suppress_missing else s.suppress_missing():
            for k in key_tuple[:-1]:
                s = s[k]
        s[key_tuple[-1]] = value

    def flatten(self, flatten_list=True):
        """Return a flattened copy of this instance.

        New keys are constructed by concatenating the (nested) keys of this
        instance into tuples.

        Opposite of the :meth:`.Settings.unflatten` method.

        If *flatten_list* is ``True``, all nested lists will be flattened as well.
        Dictionary keys are replaced with list indices in such case.

        .. code-block:: python

            >>> s = Settings()
            >>> s.a.b.c = True
            >>> print(s)
            a:
              b:
                c: 	True

            >>> s_flat = s.flatten()
            >>> print(s_flat)
            ('a', 'b', 'c'): 	True
        """
        if flatten_list:
            nested_type = (Settings, list)
            iter_type = lambda x: x.items() if isinstance(x, Settings) else enumerate(x)
        else:
            nested_type = Settings
            iter_type = Settings.items

        def _concatenate(key_ret, sequence):
            # Switch from Settings.items() to enumerate() if a list is encountered
            for k, v in iter_type(sequence):
                k = key_ret + (k,)
                if isinstance(v, nested_type) and v:
                    # Empty lists or Settings instances will return ``False``
                    _concatenate(k, v)
                else:
                    ret[k] = v

        # Changes keys into tuples
        ret = Settings()
        _concatenate((), self)
        return ret

    def unflatten(self, unflatten_list=True):
        """Return a nested copy of this instance.

        New keys are constructed by expanding the keys of this instance
        (*e.g.* tuples) into new nested |Settings| instances.

        If *unflatten_list* is ``True``, integers will be interpreted as list
        indices and are used for creating nested lists.

        Opposite of the :meth:`.Settings.flatten` method.

        .. code-block:: python

            >>> s = Settings()
            >>> s[('a', 'b', 'c')] = True
            >>> print(s)
            ('a', 'b', 'c'): 	True

            >>> s_nested = s.unflatten()
            >>> print(s_nested)
            a:
              b:
                c: 	True
        """
        ret = Settings()
        for key, value in self.items():
            s = ret
            for k1, k2 in zip(key[:-1], key[1:]):
                if not unflatten_list:
                    s = s[k1]
                    continue

                if isinstance(k2, int) and not isinstance(s[k1], list):
                    s[k1] = []
                if isinstance(k1, int):  # Apply padding to s
                    s += [Settings()] * (k1 - len(s) + 1)
                s = s[k1]
            s[key[-1]] = value

        return ret

    # =======================================================================

    def __iter__(self):
        """Iteration through keys follows lexicographical order. All keys are
        sorted as if they were strings."""
        return iter(sorted(self.keys(), key=str))

    def __missing__(self, name):
        """When requested key is not present, add it with an empty |Settings|
        instance as a value.

        This method is essential for automatic insertions in deeper levels.
        Without it things like::

            >>> s = Settings()
            >>> s.a.b.c = 12

        will not work.

        The behaviour of this method can be suppressed by initializing the
        :class:`.Settings.suppress_missing` context manager.
        """
        self[name] = Settings()
        return self[name]

    def __contains__(self, name):
        """Like regular ``__contains`__``, but ignore the case."""
        return dict.__contains__(self, self.find_case(name))

    def __getitem__(self, name):
        """Like regular ``__getitem__``, but ignore the case."""
        return dict.__getitem__(self, self.find_case(name))

    def __setitem__(self, name, value):
        """Like regular ``__setitem__``, but ignore the case and if the value
        is a dict, convert it to |Settings|."""
        if isinstance(value, dict) and not isinstance(value, Settings):
            value = Settings(value)
        dict.__setitem__(self, self.find_case(name), value)

    def __delitem__(self, name):
        """Like regular ``__detitem__``, but ignore the case."""
        return dict.__delitem__(self, self.find_case(name))

    def __getattr__(self, name):
        """If name is not a magic method, redirect it to ``__getitem__``."""
        if name.startswith("__") and name.endswith("__"):
            return dict.__getattribute__(self, name)
        return self[name]

    def __setattr__(self, name, value):
        """If name is not a magic method, redirect it to ``__setitem__``."""
        if name.startswith("__") and name.endswith("__"):
            dict.__setattr__(self, name, value)
        else:
            self[name] = value

    def __delattr__(self, name):
        """If name is not a magic method, redirect it to ``__delitem__``."""
        if name.startswith("__") and name.endswith("__"):
            dict.__delattr__(self, name)
        else:
            del self[name]

    def _str(self, indent):
        """Print contents with *indent* spaces of indentation. Recursively used
        for printing nested |Settings| instances with proper indentation."""
        ret = ""
        for key, value in self.items():
            ret += " " * indent + str(key) + ": \t"
            if isinstance(value, Settings):
                if len(value) == 0:
                    ret += "<empty Settings>\n"
                else:
                    ret += "\n" + value._str(indent + len(str(key)) + 1)
            else:  # Apply consistent indentation at every '\n' character
                indent_str = " " * (2 + indent + len(str(key))) + "\t"
                ret += textwrap.indent(str(value), indent_str)[len(indent_str) :] + "\n"
        return ret if ret else "<empty Settings>"

    def __str__(self):
        return self._str(0)

    __repr__ = __str__
    __iadd__ = soft_update
    __add__ = merge
    __copy__ = copy


class SuppressMissing(contextlib.AbstractContextManager):
    """A context manager for temporary disabling the :meth:`.Settings.__missing__`
    magic method. See :meth:`Settings.suppress_missing` for more details."""

    def __init__(self, obj: type):
        """Initialize the :class:`SuppressMissing` context manager."""
        # Ensure that obj is a class, not a class instance
        self.obj = obj if isinstance(obj, type) else type(obj)
        self.missing = obj.__missing__

    def __enter__(self):
        """Enter the :class:`SuppressMissing` context manager: delete
        :meth:`.Settings.__missing__` at the class level."""

        @wraps(self.missing)
        def __missing__(self, name):
            raise KeyError(name)

        # The __missing__ method is replaced for as long as the context manager is open
        setattr(self.obj, "__missing__", __missing__)

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the :class:`SuppressMissing` context manager: reenable
        :meth:`.Settings.__missing__` at the class level."""
        setattr(self.obj, "__missing__", self.missing)



import os
import re

import importlib_resources

from pydirac.core.periodic import Element
from pydirac.io.inputs import Mol

__all__ = [
    "basis_helper",
    "get_explicit_basis_default_basis",
    "get_custom_basis_from_ele",
    "DyallBasisHelper",
    "DyallBasis",
]


def basis_helper(filename="ANO-RCC"):
    """Custom basis set generate script.

    Parameters
    ----------
    filename : str, optional
        The name of the file containing the basis set information. Default value is "ANO-RCC".

    """
    with open(filename) as f:
        textlines = f.readlines()

    textlines = textlines[43:-1]
    res_list = []
    element_mess = []
    for line in textlines:
        first_word = line.strip().split()[0]
        if first_word == "!" and "functions" not in line:
            if len(element_mess) > 3:
                res_list.append("".join(element_mess))
                element_mess = []
            element_mess.append(line)
            continue

        elif first_word == "a":
            element_mess.append("!" + line)
            continue

        elif first_word == "H":
            nb_exp = int(line.strip().split()[1])
            tmp_str = f"f{nb_exp:4d}{0:5d}\n"
            element_mess.append(tmp_str)
            continue
        else:
            element_mess.append(line)

    res_list.append("".join(element_mess))

    if not os.path.exists("basis"):
        os.makedirs("basis")

    assert len(res_list) == 96
    for i in range(1, 97):
        symbol = Element(i).symbol
        with open(f"basis/{symbol}.dat", "w") as f:
            nb_block = res_list[i - 1].count("functions")
            f.write("LARGE EXPLICIT {} {}\n".format(nb_block, "1 " * nb_block))
            f.write(res_list[i - 1])
            f.write("FINISH\n")


def get_explicit_basis_default_basis(input_strings, fname_out=None):
    """
    Get explicit basis set from DIRAC default basis library. This involves copying
    basis information from the library to a new file that specifies the basis set using
    the 'EXPLICIT' keyword.

    This method allows for the addition of more diffuse functions to the default library
    using even-tempered methods.

    Currently, this function supports the following basis sets:
        1) dyall.acv3z
        2) dyall.acv4z
        3) dyall.cv3z
        4) dyall.cv4z

    Parameters
    ----------
    input_strings : list of str
        List of strings containing basis set information for one or more elements.
    fname_out : str or None, optional
        Name of output file to save the explicit basis set. If None, the basis set
        will not be saved to a file.

    Returns
    -------
    str
        String containing the explicit basis set information.
    """
    bs_helper = DyallBasisHelper()
    lines = input_strings.strip().split("\n")
    Z = int(float(lines[4].split()[0]))
    e_symbol = Element(Z).symbol
    basis_type = lines[6].split()[-1].split(".")[-1]

    bs = bs_helper.get_basis_set(basis_type, e_symbol)

    new_strings = []
    # lines[2] = 'c-' + lines[2]
    lines[2] = "c-dyall." + basis_type
    new_strings += lines[:6]
    nb_sym = len(bs.sym_dict)
    new_strings.append("LARGE EXPLICIT {} {}".format(nb_sym, " ".join(["1"] * nb_sym)))
    new_strings.append(bs.to_string())
    new_strings.append("FINISH")

    if fname_out is None:
        fname_out = e_symbol + "_c-dyall." + basis_type + ".mol"

    with open(fname_out, "w") as f:
        f.write("\n".join(new_strings))


def get_custom_basis_from_ele(ele_type, basis_type, fname_out=None):
    e = Element(ele_type)
    input_strings = Mol.get_mol_by_default_basis(e.symbol, e.Z, basis_type)
    get_explicit_basis_default_basis(input_strings, fname_out=fname_out)


class DyallBasisHelper:
    def __init__(self):
        pass

    def get_basis_set(self, basis_type, ele_type):
        basis_type = basis_type.lower()
        if basis_type not in ["acv3z", "acv4z", "cv3z", "cv4z"]:
            raise RuntimeError('basis_type must be one of "acv3z", "acv4z", ' '"cv3z", and "cv4z"')

        fname = str(importlib_resources.files("pydirac.data.basisset") / f"dyall.{basis_type}")
        with open(fname) as f:
            lines = f.readlines()

        comment = []
        i_start = 0
        for i, line in enumerate(lines):
            match = re.match(r"^\$", line)
            if match:
                comment.append(line.rstrip())
                continue
            else:
                i_start = i
                break

        if i_start == 0:
            raise RuntimeError("Something wrong in the basis set file")

        i_start -= 1
        e = Element(ele_type)

        lines = lines[i_start:]
        for j, line in enumerate(lines):
            match = re.match(rf"^\$ {e.symbol}\s*\n?$", line)
            # if line.startswith('$ {}'.format(e.symbol)):
            if match:
                j_start = j
                break
        else:
            raise RuntimeError(
                "This basis set does not support element " "type {}".format(e.symbol)
            )

        lines = lines[j_start:]
        tmp_strings = []
        for line in lines:
            match = re.match(r"^\s*\n?$", line)
            # if len(line.strip()) == 0:
            if match:
                break
            else:
                tmp_strings.append(line.rstrip())

        dyall_basis = DyallBasis.from_string("\n".join(tmp_strings))
        return dyall_basis


class DyallBasis:
    def __init__(self, ele_Z, sym_dict):
        self.e = Element(ele_Z)
        self.symbol = self.e.symbol
        self.Z = self.e.Z
        self.sym_dict = sym_dict

    @classmethod
    def from_string(cls, bs_text):
        bs_info = bs_text.strip().split("\n")
        # the first line is a comment that include element symbol info
        symbol = bs_info[0].split()[-1]
        Z = int(bs_info[1].split()[-1])
        assert Element(symbol).Z == Z
        sym_dict = {}
        curr_sym = None
        for line in bs_info[2:]:
            if line.startswith("$"):
                sym = line.split()[1]
                if sym not in sym_dict:
                    sym_dict[sym] = []
                curr_sym = sym
            elif len(line.strip().split()) > 1:
                continue
            else:
                sym_dict[curr_sym].append(line.rstrip())
        return DyallBasis(Z, sym_dict)

    def to_string(self):
        tmp_strings = []
        # tmp_strings.append('$ {}'.format(self.symbol))
        # tmp_strings.append('a {}'.format(self.Z))
        for sym in "spdfghi":
            if sym in self.sym_dict:
                tmp_strings.append(f"$ {sym} functions")
                nb_exp = len(self.sym_dict[sym])
                tmp_strings.append(f"f{nb_exp:>4}    0    0")
                for exp in self.sym_dict[sym]:
                    tmp_strings.append(f"    {float(exp):10.8E}")
        return "\n".join(tmp_strings)

    def __str__(self):
        return self.to_string()


if __name__ == "__main__":
    # basis_helper(filename='dirac_basis')
    bs_helper = DyallBasisHelper()
    bs = bs_helper.get_basis_set("acv4z", "Kr")
    print(bs)
    pass
